from functools import partial
from multiprocessing import Pool
from itertools import starmap

import numpy as np
from bilby.core.prior import Interped
from threadpoolctl import threadpool_limits


def interpolated_sample_and_log_prob_multi(
    sample_points, values, num_processes: int = 1
):
    """
    Given a distribution discretized on a grid, return a sample and the log prob from an
    interpolated distribution. Wraps the bilby.core.prior.Interped class. Works with
    multiprocessing.

    Parameters
    ----------
    sample_points : np.ndarray, shape (N)
        x values for samples
    values : np.ndarray, shape (B, N)
        y values for samples. The distributions do not have to be initially
        normalized, although the final log_probs will be. B = batch dimension.
    num_processes : int
        Number of parallel processes to use.

    Returns
    -------
    (np.ndarray, np.ndarray) : sample and log_prob arrays, each of length B
    """
    with threadpool_limits(limits=1, user_api="blas"):
        data_generator = iter(values)
        task_fun = partial(interpolated_sample_and_log_prob, sample_points)
        if num_processes > 1:
            with Pool(processes=num_processes) as pool:
                result_list = pool.map(task_fun, data_generator)
        else:
            result_list = list(map(task_fun, data_generator))
    sample, log_prob = np.array(result_list).T
    return sample, log_prob


def interpolated_sample_and_log_prob(sample_points, values):
    """
    Given a distribution discretized on a grid, return a sample and the log prob from an
    interpolated distribution. Wraps the bilby.core.prior.Interped class.

    Parameters
    ----------
    sample_points : np.ndarray
        x values for samples
    values : np.ndarray
        y values for samples. The distribution does not have to be initially
        normalized, although the final log_prob will be.

    Returns
    -------
    (float, float) : sample and log_prob
    """
    interp = Interped(sample_points, values)
    sample = interp.sample()
    log_prob = interp.ln_prob(sample)
    return sample, log_prob


def interpolated_log_prob_multi(
    sample_points, values, evaluation_points, num_processes: int = 1
):
    """
    Given a distribution discretized on a grid, the log prob at a specific point
    using an interpolated distribution. Wraps the bilby.core.prior.Interped class.
    Works with multiprocessing.

    Parameters
    ----------
    sample_points : np.ndarray, shape (N)
        x values for samples
    values : np.ndarray, shape (B, N)
        y values for samples. The distributions do not have to be initially
        normalized, although the final log_probs will be. B = batch dimension.
    evaluation_points : np.ndarray, shape (B)
        x values at which to evaluate log_prob.
    num_processes : int
        Number of parallel processes to use.

    Returns
    -------
    (np.ndarray, np.ndarray) : sample and log_prob arrays, each of length B
    """
    with threadpool_limits(limits=1, user_api="blas"):
        data_generator = zip(iter(values), iter(evaluation_points))
        task_fun = partial(interpolated_log_prob, sample_points)
        if num_processes > 1:
            with Pool(processes=num_processes) as pool:
                result_list = pool.starmap(task_fun, data_generator)
        else:
            result_list = list(starmap(task_fun, data_generator))
    return np.array(result_list)


def interpolated_log_prob(sample_points, values, evaluation_point):
    """
    Given a distribution discretized on a grid, return a sample and the log prob from an
    interpolated distribution. Wraps the bilby.core.prior.Interped class.

    Parameters
    ----------
    sample_points : np.ndarray
        x values for samples
    values : np.ndarray
        y values for samples. The distribution does not have to be initially
        normalized, although the final log_prob will be.
    evaluation_point : float
        x value at which to evaluate log_prob.

    Returns
    -------
    float : log_prob
    """
    interp = Interped(sample_points, values)
    return interp.ln_prob(evaluation_point)



from scipy.stats import kstest, gaussian_kde
import numpy as np


class IterationTracker:
    def __init__(self, data=None, store_data=False):
        self.data = data
        self.ks_result = None
        self.store_data = store_data

    def update(self, new_data):
        """
        Append new_data to self.data.

        Parameters
        ----------
        new_data: dict
            dict with numpy arrays to append to data

        Returns
        -------

        """
        if self.data is None:
            self.data = {k: v.copy()[None, :] for k, v in new_data.items()}

        else:
            x = {k: v[-1, :] for k, v in self.data.items()}
            y = new_data
            # get ks
            statistic, pvalue = [], []
            for k in y.keys():
                N = len(x[k]) // 2
                ks_result = kstest(x[k][:N], y[k][N:])
                statistic.append(ks_result[0])
                pvalue.append(ks_result[1])
            self.ks_result = {"statistics": statistic, "pvalue": pvalue}

            if not self.store_data:
                self.data = {k: v.copy()[None, :] for k, v in y.items()}
            else:
                self.data = {
                k: np.concatenate((v, y[k][None, :]), axis=0)
                for k, v in self.data.items()
            }

    @property
    def pvalue_min(self):
        if self.ks_result is None:
            return -np.inf
        else:
            return min(self.ks_result["pvalue"])

    # def remove_outliers(self, x):
    #     xc = np.concatenate([v[None, :] for v in self.x.values()], axis=0)
    #     yc = np.concatenate([v[None, :] for v in y.values()], axis=0)
    #     gaussian_kde(yc).logpdf(yc)



import pytest
import h5py
import numpy as np

from dingo.gw.domains import build_domain
from dingo.gw.noise.asd_dataset import ASDDataset


@pytest.fixture
def noise_dataset():
    dataset_path = "./asds_toydataset.hdf5"

    domain_dict = {
        "type": "FrequencyDomain",
        "f_min": 0.0,
        "f_max": 100.0,
        "delta_f": 1.0,
    }
    domain = build_domain(domain_dict)
    domain_dict = domain.domain_dict
    ifos_num_asds = {"H1": 10, "L1": 8, "V1": 5}

    f = h5py.File(dataset_path, "w")
    gps_times = f.create_group("gps_times")
    grp = f.create_group("asds")
    settings = {"domain_dict": domain.domain_dict}

    for ifo, num_asds in ifos_num_asds.items():
        asds = np.sin(np.outer(np.arange(num_asds) / 100, domain()))
        grp.create_dataset(ifo, data=asds)
        gps_times[ifo] = np.arange(num_asds)

    f.attrs["settings"] = str(settings)
    f.close()

    return dataset_path, domain_dict, ifos_num_asds


def test_noise_dataset_load(noise_dataset):
    dataset_path, domain_dict, ifos_num_asds = noise_dataset

    asd_dataset = ASDDataset(dataset_path)
    # check domain
    assert asd_dataset.domain.domain_dict == domain_dict
    # check that the dataset has the correct ifos and corresponding datasets
    assert asd_dataset.asds.keys() == ifos_num_asds.keys()
    for ifo, n in ifos_num_asds.items():
        assert asd_dataset.asds[ifo].shape == (n, len(asd_dataset.domain))
    # check that sampling work correctly
    asd_samples = asd_dataset.sample_random_asds()
    assert asd_samples.keys() == ifos_num_asds.keys()
    for _, asd in asd_samples.items():
        assert len(asd) == len(asd_dataset.domain)

    # repeat tests for ASDDataset with only a subset of detectors
    ifos = ["H1", "V1"]
    asd_dataset = ASDDataset(dataset_path, ifos=ifos)
    # check domain
    assert asd_dataset.domain.domain_dict == domain_dict
    # check that the dataset has the correct ifos and corresponding datasets
    assert set(asd_dataset.asds.keys()) == set(ifos)
    for ifo in ifos:
        assert asd_dataset.asds[ifo].shape == (
            ifos_num_asds[ifo],
            len(asd_dataset.domain),
        )
    # check that sampling work correctly
    asd_samples = asd_dataset.sample_random_asds()
    assert set(asd_samples.keys()) == set(ifos)
    for _, asd in asd_samples.items():
        assert len(asd) == len(asd_dataset.domain)


def test_noise_dataset_update_domain(noise_dataset):
    dataset_path, domain_dict, ifos_num_asds = noise_dataset

    asd_dataset = ASDDataset(dataset_path)
    asd_samples = asd_dataset.sample_random_asds()

    delta_f = asd_dataset.domain.delta_f
    f_min = asd_dataset.domain.f_min
    f_max = asd_dataset.domain.f_max
    f_min_new = 20
    f_max_new = 40
    domain_update = {"f_min": f_min_new, "f_max": f_max_new}
    asd_dataset_truncated = ASDDataset(dataset_path, domain_update=domain_update)
    asd_samples_truncated = asd_dataset_truncated.sample_random_asds()

    domain_ref = build_domain(domain_dict)
    domain_ref.update(domain_update)
    nf_tr = len(domain_ref)
    nf = len(asd_dataset.domain)

    for ifo, n in ifos_num_asds.items():
        asd_data_full = asd_dataset.asds[ifo]
        asd_data_full_truncated = asd_dataset_truncated.asds[ifo]
        print("cp")
        # check that asd data and asd samples have the correct dimensions
        assert asd_data_full.shape == (n, nf)
        assert len(asd_samples[ifo]) == nf
        assert asd_data_full_truncated.shape == (n, nf_tr)
        assert len(asd_samples_truncated[ifo]) == nf_tr
        # check that the truncation worked properly
        assert np.all(
            asd_data_full[
                :, round(f_min_new / delta_f) : round(f_max_new / delta_f) + 1
            ]
            == asd_data_full_truncated[:, asd_dataset_truncated.domain.min_idx :]
        )
        assert not np.all(
            asd_data_full[:, round(f_min / delta_f) : round(f_min_new / delta_f)]
            == asd_data_full_truncated[
                :, round(f_min / delta_f) : asd_dataset_truncated.domain.min_idx
            ]
        )
        assert np.all(
            asd_data_full_truncated[:, : asd_dataset_truncated.domain.min_idx] == 1.0
        )



import os

from bilby_pipe.job_creation.nodes import AnalysisNode

from dingo.pipe.utils import _strip_unwanted_submission_keys


class ImportanceSamplingNode(AnalysisNode):
    def __init__(self, inputs, sampling_node, generation_node, parallel_idx, dag):
        super(AnalysisNode, self).__init__(inputs)
        self.dag = dag
        self.sampling_node = sampling_node
        self.generation_node = generation_node
        self.parallel_idx = parallel_idx
        self.request_cpus = inputs.request_cpus_importance_sampling

        data_label = sampling_node.job_name
        base_name = data_label.replace("sampling", "importance_sampling")
        self.base_job_name = base_name
        if parallel_idx != "":
            self.job_name = f"{base_name}_{parallel_idx}"
        else:
            self.job_name = base_name
        self.label = self.job_name

        proposal_samples_file = os.path.join(
            self.inputs.result_directory,
            self.label.replace("importance_sampling", "sampling") + ".hdf5",
        )

        self.setup_arguments()

        # if self.inputs.transfer_files or self.inputs.osg:
        #     data_dump_file = generation_node.data_dump_file
        #     input_files_to_transfer = [
        #         str(data_dump_file),
        #         str(self.inputs.complete_ini_file),
        #     ]
        #     self.extra_lines.extend(
        #         self._condor_file_transfer_lines(
        #             input_files_to_transfer,
        #             [self._relative_topdir(self.inputs.outdir, self.inputs.initialdir)],
        #         )
        #     )
        #     self.arguments.add("outdir", os.path.relpath(self.inputs.outdir))

        # Add extra arguments for dingo
        self.arguments.add("label", self.label)
        self.arguments.add("proposal-samples-file", proposal_samples_file)
        self.arguments.add("event-data-file", generation_node.event_data_file)

        self.extra_lines.extend(self._checkpoint_submit_lines())
        # if self.request_cpus > 1:
        #     self.extra_lines.extend(['environment = "OMP_NUM_THREADS=1"'])

        self.process_node()

        # We need both of these as parents because importance sampling can in principle
        # use different data than sampling. In that case, the generation node will not
        # be a parent of the sampling node.
        self.job.add_parent(sampling_node.job)
        self.job.add_parent(generation_node.job)

        if self.inputs.simple_submission:
            _strip_unwanted_submission_keys(self.job)

    @property
    def executable(self):
        return self._get_executable_path("dingo_pipe_importance_sampling")

    @property
    def result_file(self):
        return f"{self.inputs.result_directory}/{self.job_name}.hdf5"



import numpy as np
import torch
from bilby.core.prior import PriorDict
from abc import ABC, abstractmethod


class GNPEBase(ABC):
    """
    A base class for Group Equivariant Neural Posterior Estimation [1].

    This implements GNPE for *approximate* equivariances. For exact equivariances,
    additional processing should be implemented within a subclass.

    [1]: https://arxiv.org/abs/2111.13139
    """

    def __init__(self, kernel_dict, operators):
        self.kernel = PriorDict(kernel_dict)
        self.operators = operators
        self.proxy_list = [k + "_proxy" for k in kernel_dict.keys()]
        self.context_parameters = self.proxy_list.copy()
        self.input_parameter_names = list(self.kernel.keys())

    @abstractmethod
    def __call__(self, input_sample):
        pass

    def sample_proxies(self, input_parameters):
        """
        Given input parameters, perturbs based on the
        kernel to produce "proxy" ("hatted") parameters, i.e., samples

            \hat g ~ p(\hat g | g).

        Typically the GNPE NDE will be conditioned on \hat g. Furthermore, these proxy
        parameters will be used to transform the data to simplify it.

        Parameters:
        -----------
        input_parameters : dict
            Initial parameter values to be perturbed. dict values can be either floats
            (for training) or torch Tensors (for inference).

        Returns
        -------
        A dict of proxy parameters.
        """
        proxies = {}
        for k in self.kernel:
            if k not in input_parameters:
                raise KeyError(
                    f"Input parameters are missing key {k} required for GNPE."
                )
            g = input_parameters[k]
            g_hat = self.perturb(g, k)
            proxies[k + "_proxy"] = g_hat
        return proxies

    def perturb(self, g, k):
        """
        Generate proxy variables based on initial parameter values.

        Parameters
        ----------
        g : Union[np.float64, float, torch.Tensor]
            Initial parameter values
        k : str
            Parameter name. This is used to identify the group binary operator.

        Returns
        -------
        Proxy variables in the same format as g.
        """
        # First we sample from the kernel, ensuring the correct data type,
        # and accounting for possible batching.
        #
        # Batching is implemented only for torch Tensors (expected at inference time),
        # whereas un-batched data in float form is expected during training.
        if type(g) == torch.Tensor:
            epsilon = self.kernel[k].sample(len(g))
            epsilon = torch.tensor(epsilon, dtype=g.dtype, device=g.device)
        elif type(g) == np.float64 or type(g) == float:
            epsilon = self.kernel[k].sample()
        else:
            raise NotImplementedError(f"Unsupported data type {type(g)}.")

        return self.multiply(g, epsilon, k)

    def multiply(self, a, b, k):
        op = self.operators[k]
        if op == "+":
            return a + b
        elif op == "x":
            return a * b
        else:
            raise NotImplementedError(
                f"Unsupported group multiplication operator: {op}"
            )

    def inverse(self, a, k):
        op = self.operators[k]
        if op == "+":
            return -a
        elif op == "x":
            return 1 / a
        else:
            raise NotImplementedError(
                f"Unsupported group multiplication operator: {op}"
            )


class GNPECoalescenceTimes(GNPEBase):
    """
    GNPE [1] Transformation for detector coalescence times.

    For each of the detector coalescence times, a proxy is generated by adding a
    perturbation epsilon from the GNPE kernel to the true detector time. This proxy is
    subtracted from the detector time, such that the overall time shift only amounts to
    -epsilon in training. This standardizes the input data to the inference network,
    since the applied time shifts are always restricted to the range of the kernel.

    To preserve information at inference time, conditioning of the inference network on
    the proxies is required. To that end, the proxies are stored in sample[
    'gnpe_proxies'].

    We can enforce an exact equivariance under global time translations, by subtracting
    one proxy (by convention: the first one, usually for H1 ifo) from all other
    proxies, and from the geocent time, see [1]. This is enabled with the flag
    exact_global_equivariance.

    Note that this transform does not modify the data itself. It only determines the
    amount by which to time-shift the data.

    [1]: arxiv.org/abs/2111.13139
    """

    def __init__(
        self, ifo_list, kernel, exact_global_equivariance=True, inference=False
    ):
        """
        Parameters
        ----------
        ifo_list : bilby.gw.detector.InterferometerList
            List of interferometers.
        kernel : str
            Defines a Bilby prior, to be used for all interferometers.
        exact_global_equivariance : bool = True
            Whether to impose the exact global time translation symmetry.
        inference : bool = False
            Whether to use inference or training mode.
        """
        self.ifo_time_labels = [ifo.name + "_time" for ifo in ifo_list]
        kernel_dict = {k: kernel for k in self.ifo_time_labels}
        operators = {k: "+" for k in self.ifo_time_labels}
        super().__init__(kernel_dict, operators)

        self.inference = inference
        self.exact_global_equivariance = exact_global_equivariance
        if self.exact_global_equivariance:
            # GNPE networks are conditioned on proxy variables relative to the
            # "preferred" proxy (typically H1). We give these a different name so that we
            # can keep track separately of the un-shifted proxies.
            del self.context_parameters[0]
            self.context_parameters = [p + "_relative" for p in self.context_parameters]

    def __call__(self, input_sample):
        sample = input_sample.copy()
        extrinsic_parameters = sample["extrinsic_parameters"].copy()

        # If proxies already exist, use them. Otherwise, sample them. Proxies may
        # already exist if provided by an unconditional initialization network when
        # attempting to recover the density from GNPE samples.

        # TODO: Reimplement in GNPEBase.sample_proxies().
        if set(self.proxy_list).issubset(extrinsic_parameters.keys()):
            new_parameters = {p: extrinsic_parameters[p] for p in self.proxy_list}
        else:
            new_parameters = self.sample_proxies(extrinsic_parameters)

        # If we are in training mode, we assume that the time shifting due to different
        # arrival times of the signal in individual detectors has not yet been applied
        # to the data; instead the arrival times are stored in extrinsic_parameters.
        # Hence we subtract off the proxy times from these arrival times, so that time
        # shifting of the data only has to be done once.

        if not self.inference:
            for k in self.ifo_time_labels:
                new_parameters[k] = (
                    -new_parameters[k + "_proxy"] + extrinsic_parameters[k]
                )
        # In inference mode, the data are only time shifted by minus the proxy.
        else:
            for k in self.ifo_time_labels:
                new_parameters[k] = -new_parameters[k + "_proxy"]

        # If we are imposing the global time shift symmetry, then we treat the first
        # proxy as "preferred", in the sense that it defines the global time shift.
        # This symmetry is enforced as follows:
        #
        #    1) Do not explicitly condition the model on the preferred proxy
        #    2) Subtract the preferred proxy from geocent_time (assumed to be a regression
        #    parameter). Note that this must be undone at inference time.
        #    3) Subtract the preferred proxy from the remaining proxies. These remaining
        #    proxies then define time shifts relative to the global time shift.
        #
        # Imposing the global time shift does not impact the transformation of the
        # data: we do not change the values of the true detector coalescence times
        # stored in extrinsic_parameters, only the proxies.

        if self.exact_global_equivariance:
            dt = new_parameters[self.ifo_time_labels[0] + "_proxy"]
            if not self.inference:
                if "geocent_time" not in extrinsic_parameters:
                    raise KeyError(
                        "geocent_time should be in extrinsic_parameters at "
                        "this point during training."
                    )
                new_parameters["geocent_time"] = (
                    extrinsic_parameters["geocent_time"] - dt
                )
            else:
                new_parameters["geocent_time"] = -dt
            for k in self.ifo_time_labels[1:]:
                new_parameters[k + "_proxy_relative"] = (
                    new_parameters[k + "_proxy"] - dt
                )

        extrinsic_parameters.update(new_parameters)
        sample["extrinsic_parameters"] = extrinsic_parameters
        return sample



"""
TODO: Docstring
"""

from typing import Callable
import torch
import dingo.core.utils as utils
from torch.utils.data import Dataset
import os
import time
import numpy as np
from threadpoolctl import threadpool_limits
import dingo.core.utils.trainutils
import math
import h5py
import json
from collections import OrderedDict

from dingo.core.nn.nsf import (
    create_nsf_with_rb_projection_embedding_net,
    create_nsf_wrapped,
)
from dingo.core.utils.misc import get_version


class PosteriorModel:
    """
    TODO: Docstring

    Methods
    -------

    initialize_model:
        initialize the NDE (including embedding net) as posterior model
    initialize_training:
        initialize for training, that includes storing the epoch, building
        an optimizer and a learning rate scheduler
    save_model:
        save the model, including all information required to rebuild it,
        except for the builder function
    load_model:
        load and build a model from a file
    train_model:
        train the model
    inference:
        perform inference
    """

    def __init__(
        self,
        model_filename: str = None,
        metadata: dict = None,
        initial_weights: dict = None,
        device: str = "cuda",
        load_training_info: bool = True,
    ):
        """

        Parameters
        ----------

        model_builder: Callable
            builder function for the model,
            self.model = model_builder(**model_kwargs)
        model_kwargs: dict = None
            kwargs for for the model,
            self.model = model_builder(**model_kwargs)
        model_filename: str = None
            path to filename of loaded model
        optimizer_kwargs: dict = None
            kwargs for optimizer
        scheduler_kwargs: dict = None
            kwargs for scheduler
        init_for_training: bool = False
            flag whether initialization for training (e.g., optimizer) required
        metadata: dict = None
            dict with metadata, used to save dataset_settings and train_settings
        """
        self.version = f"dingo={get_version()}"  # dingo version

        self.optimizer_kwargs = None
        self.model_kwargs = None
        self.scheduler_kwargs = None
        self.initial_weights = initial_weights

        self.metadata = metadata
        if self.metadata is not None:
            self.model_kwargs = self.metadata["train_settings"]["model"]
            # Expect self.optimizer_settings and self.scheduler_settings to be set
            # separately, and before calling initialize_optimizer_and_scheduler().

        self.epoch = 0
        self.optimizer = None
        self.scheduler = None
        self.context = None
        self.event_metadata = None

        # build model
        if model_filename is not None:
            self.load_model(
                model_filename, load_training_info=load_training_info, device=device
            )
        else:
            self.initialize_model()
            self.model_to_device(device)

    def model_to_device(self, device):
        """
        Put model to device, and set self.device accordingly.
        """
        if device not in ("cpu", "cuda"):
            raise ValueError(f"Device should be either cpu or cuda, got {device}.")
        self.device = torch.device(device)
        # Commented below so that code runs on first cuda device in the case of multiple.
        # if device == 'cuda' and torch.cuda.device_count() > 1:
        #     print("Using", torch.cuda.device_count(), "GPUs.")
        #     raise NotImplementedError('This needs testing!')
        #     # dim = 0 [512, ...] -> [256, ...], [256, ...] on 2 GPUs
        #     self.model = torch.nn.DataParallel(self.model)
        print(f"Putting posterior model to device {self.device}.")
        self.model.to(self.device)

    def initialize_model(self):
        """
        Initialize a model for the posterior by calling the
        self.model_builder with self.model_kwargs.

        """
        model_builder = get_model_callable(self.model_kwargs["type"])
        model_kwargs = {k: v for k, v in self.model_kwargs.items() if k != "type"}
        if self.initial_weights is not None:
            model_kwargs["initial_weights"] = self.initial_weights
        self.model = model_builder(**model_kwargs)

    def initialize_optimizer_and_scheduler(self):
        """
        Initializes the optimizer and scheduler with self.optimizer_kwargs
        and self.scheduler_kwargs, respectively.
        """
        if self.optimizer_kwargs is not None:
            self.optimizer = utils.get_optimizer_from_kwargs(
                self.model.parameters(), **self.optimizer_kwargs
            )
        if self.scheduler_kwargs is not None:
            self.scheduler = utils.get_scheduler_from_kwargs(
                self.optimizer, **self.scheduler_kwargs
            )

    def save_model(
        self,
        model_filename: str,
        save_training_info: bool = True,
    ):
        """
        Save the posterior model to the disk.

        Parameters
        ----------
        model_filename: str
            filename for saving the model
        save_training_info: bool
            specifies whether information required to proceed with training is
            saved, e.g. optimizer state dict

        """
        model_dict = {
            "model_kwargs": self.model_kwargs,
            "model_state_dict": self.model.state_dict(),
            "epoch": self.epoch,
            "version": self.version,
        }

        if self.metadata is not None:
            model_dict["metadata"] = self.metadata

        if self.context is not None:
            model_dict["context"] = self.context

        if self.event_metadata is not None:
            model_dict["event_metadata"] = self.event_metadata

        if save_training_info:
            model_dict["optimizer_kwargs"] = self.optimizer_kwargs
            model_dict["scheduler_kwargs"] = self.scheduler_kwargs
            if self.optimizer is not None:
                model_dict["optimizer_state_dict"] = self.optimizer.state_dict()
            if self.scheduler is not None:
                model_dict["scheduler_state_dict"] = self.scheduler.state_dict()
            # TODO

        torch.save(model_dict, model_filename)


    def _load_model_from_hdf5(
        self,
        model_filename: str
    ):
        """
        Helper function to load a trained model that has been
        saved in HDF5 format using `dingo_pt_to_hdf5`.

        Parameters
        ----------
        model_filename: str
            path to saved model; must have extension '.hdf5'

        Returns
        -------
        d: dict
            A stripped down version of the dict saved by torch.save()
            Specifically, it does not include 'optimizer_state_dict'
            to save space at inference time.
        """
        d = {}
        with h5py.File(model_filename, 'r') as fp:
            model_basename = os.path.basename(model_filename)
            if fp.attrs['CANONICAL_FILE_BASENAME'] != model_basename:
                raise ValueError('HDF5 attribute CANONICAL_FILE_BASENAME differs from model name',
                        model_basename)

            # Load small nested dicts from json
            for k, v in fp['serialized_dicts'].items():
                d[k] = json.loads(v[()])

            # Load model weights
            model_state_dict = OrderedDict()
            for k, v in fp['model_weights'].items():
                model_state_dict[k] = torch.from_numpy(np.array(v, dtype=np.float32))
            d['model_state_dict'] = model_state_dict

        return d


    def load_model(
        self,
        model_filename: str,
        load_training_info: bool = True,
        device: str = "cuda",
    ):
        """
        Load a posterior model from the disk.

        Parameters
        ----------
        model_filename: str
            path to saved model
        load_training_info: bool #TODO: load information for training
            specifies whether information required to proceed with training is
            loaded, e.g. optimizer state dict
        """
        # Make sure that when the model is loaded, the torch tensors are put on the
        # device indicated in the saved metadata. External routines run on a cpu
        # machine may have moved the model from 'cuda' to 'cpu'.
        ext = os.path.splitext(model_filename)[-1]
        if ext == '.pt':
            d = torch.load(model_filename, map_location=device)
        elif ext == '.hdf5':
            d = self._load_model_from_hdf5(model_filename)
        else:
            raise ValueError('Models should be ether in .pt or .hdf5 format.')

        self.version = d.get("version")

        self.model_kwargs = d["model_kwargs"]
        self.initialize_model()
        self.model.load_state_dict(d["model_state_dict"])

        self.epoch = d["epoch"]

        self.metadata = d["metadata"]

        if "context" in d:
            self.context = d["context"]

        if "event_metadata" in d:
            self.event_metadata = d["event_metadata"]

        self.model_to_device(device)

        if load_training_info:
            if "optimizer_kwargs" in d:
                self.optimizer_kwargs = d["optimizer_kwargs"]
            if "scheduler_kwargs" in d:
                self.scheduler_kwargs = d["scheduler_kwargs"]
            # initialize optimizer and scheduler
            self.initialize_optimizer_and_scheduler()
            # load optimizer and scheduler state dict
            if "optimizer_state_dict" in d:
                self.optimizer.load_state_dict(d["optimizer_state_dict"])
            if "scheduler_state_dict" in d:
                self.scheduler.load_state_dict(d["scheduler_state_dict"])
        else:
            # put model in evaluation mode
            self.model.eval()

    def train(
        self,
        train_loader: torch.utils.data.DataLoader,
        test_loader: torch.utils.data.DataLoader,
        train_dir: str,
        runtime_limits: object = None,
        checkpoint_epochs: int = None,
        use_wandb=False,
        test_only=False,
    ):
        """

        Parameters
        ----------
        train_loader
        test_loader
        train_dir
        runtime_limits
        checkpoint_epochs
        use_wandb
        test_only: bool = False
            if True, training is skipped

        Returns
        -------

        """
        if test_only:
            test_loss = test_epoch(self, test_loader)
            print(f"test loss: {test_loss:.3f}")

        else:
            while not runtime_limits.limits_exceeded(self.epoch):
                self.epoch += 1

                # Training
                lr = utils.get_lr(self.optimizer)
                with threadpool_limits(limits=1, user_api="blas"):
                    print(f"\nStart training epoch {self.epoch} with lr {lr}")
                    time_start = time.time()
                    train_loss = train_epoch(self, train_loader)
                    train_time = time.time() - time_start

                    print(
                        "Done. This took {:2.0f}:{:2.0f} min.".format(
                            *divmod(train_time, 60)
                        )
                    )

                    # Testing
                    print(f"Start testing epoch {self.epoch}")
                    time_start = time.time()
                    test_loss = test_epoch(self, test_loader)
                    test_time = time.time() - time_start

                    print(
                        "Done. This took {:2.0f}:{:2.0f} min.".format(
                            *divmod(time.time() - time_start, 60)
                        )
                    )

                # scheduler step for learning rate
                utils.perform_scheduler_step(self.scheduler, test_loss)

                # write history and save model
                utils.write_history(train_dir, self.epoch, train_loss, test_loss, lr)
                utils.save_model(self, train_dir, checkpoint_epochs=checkpoint_epochs)
                if use_wandb:
                    try:
                        import wandb
                        wandb.define_metric("epoch")
                        wandb.define_metric("*", step_metric="epoch")
                        wandb.log(
                            {
                                "epoch": self.epoch,
                                "learning_rate": lr[0],
                                "train_loss": train_loss,
                                "test_loss": test_loss,
                                "train_time": train_time,
                                "test_time": test_time,
                            }
                        )
                    except ImportError:
                        print("wandb not installed. Skipping logging to wandb.")

                print(f"Finished training epoch {self.epoch}.\n")

    def sample(
        self,
        *x,
        batch_size=None,
        get_log_prob=False,
    ):
        """
        Sample from posterior model, conditioned on context x. x is expected to have a
        batch dimension, i.e., to obtain N samples with additional context requires
        x = x_.expand(N, *x_.shape).

        This method takes care of the batching, makes sure that self.model is in
        evaluation mode and disables gradient computation.

        Parameters
        ----------
        *x:
            input context to the neural network; has potentially multiple elements for,
            e.g., gnpe proxies
        batch_size: int = None
            batch size for sampling
        get_log_prob: bool = False
            if True, also return log probability along with the samples

        Returns
        -------
        samples: torch.Tensor
            samples from posterior model
        """
        self.model.eval()
        with torch.no_grad():
            if batch_size is None:
                samples = self.model.sample(*x)
                if get_log_prob:
                    log_prob = self.model.log_prob(samples, *x)
            else:
                samples = []
                if get_log_prob:
                    log_prob = []
                num_batches = math.ceil(len(x[0]) / batch_size)
                for idx_batch in range(num_batches):
                    lower, upper = idx_batch * batch_size, (idx_batch + 1) * batch_size
                    x_batch = [xi[lower:upper] for xi in x]
                    samples.append(self.model.sample(*x_batch, num_samples=1))
                    if get_log_prob:
                        log_prob.append(self.model.log_prob(samples[-1], *x_batch))
                samples = torch.cat(samples, dim=0)
                if get_log_prob:
                    log_prob = torch.cat(log_prob, dim=0)
        if not get_log_prob:
            return samples
        else:
            return samples, log_prob


def get_model_callable(model_type: str):
    if model_type == "nsf+embedding":
        return create_nsf_with_rb_projection_embedding_net
    elif model_type == "nsf":
        return create_nsf_wrapped
    else:
        raise KeyError("Invalid model type.")


def train_epoch(pm, dataloader):
    pm.model.train()
    loss_info = dingo.core.utils.trainutils.LossInfo(
        pm.epoch,
        len(dataloader.dataset),
        dataloader.batch_size,
        mode="Train",
        print_freq=1,
    )

    for batch_idx, data in enumerate(dataloader):
        loss_info.update_timer()
        pm.optimizer.zero_grad()
        # data to device
        data = [d.to(pm.device, non_blocking=True) for d in data]
        # compute loss
        loss = -pm.model(data[0], *data[1:]).mean()
        # backward pass and optimizer step
        loss.backward()
        pm.optimizer.step()
        # update loss for history and logging
        loss_info.update(loss.detach().item(), len(data[0]))
        loss_info.print_info(batch_idx)

    return loss_info.get_avg()


def test_epoch(pm, dataloader):
    with torch.no_grad():
        pm.model.eval()
        loss_info = dingo.core.utils.trainutils.LossInfo(
            pm.epoch,
            len(dataloader.dataset),
            dataloader.batch_size,
            mode="Test",
            print_freq=1,
        )

        for batch_idx, data in enumerate(dataloader):
            loss_info.update_timer()
            # data to device
            data = [d.to(pm.device, non_blocking=True) for d in data]
            # compute loss
            loss = -pm.model(data[0], *data[1:]).mean()
            # update loss for history and logging
            loss_info.update(loss.item(), len(data[0]))
            loss_info.print_info(batch_idx)

        return loss_info.get_avg()



import torch
import argparse


def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument('--checkpoint', type=str, required=True)
    parser.add_argument('--out_file', type=str, required=True)
    return parser.parse_args()


def main():
    args = parse_args()

    # Typically training is done on the GPU, so the model could be saved on a GPU
    # device. Since this routine may be run on a CPU machine, allow for a remap of the
    # torch tensors.
    if torch.cuda.is_available():
        d = torch.load(args.checkpoint)
    else:
        d = torch.load(args.checkpoint, map_location=torch.device('cpu'))

    data_settings = d['metadata']['train_settings']['data']
    f_min, f_max = data_settings['conditioning']['frequency_range']
    del data_settings['conditioning']['frequency_range']

    data_settings['domain_update'] = {'f_min': f_min, 'f_max': f_max}

    torch.save(d, args.out_file)


if __name__ == "__main__":
    main()



from functools import partial
from multiprocessing import Pool
from math import isclose

import numpy as np
import astropy.units as u
from typing import Dict, List, Tuple, Union
from numbers import Number
import warnings
import lal
import lalsimulation as LS
import pandas as pd

try:
    from lalsimulation.gwsignal.core import waveform as gws_wfm
    from lalsimulation.gwsignal.models import (
        gwsignal_get_waveform_generator as new_interface_get_waveform_generator,
    )
except ImportError:
    pass

from bilby.gw.conversion import (
    convert_to_lal_binary_black_hole_parameters,
    bilby_to_lalsimulation_spins,
)
import dingo.gw.waveform_generator.wfg_utils as wfg_utils
import dingo.gw.waveform_generator.frame_utils as frame_utils
from dingo.gw.domains import Domain, FrequencyDomain, TimeDomain


class WaveformGenerator:
    """Generate polarizations using LALSimulation routines in the specified domain for a
    single GW coalescence given a set of waveform parameters.
    """

    def __init__(
        self,
        approximant: str,
        domain: Domain,
        f_ref: float,
        f_start: float = None,
        mode_list: List[Tuple] = None,
        transform=None,
        spin_conversion_phase=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        approximant : str
            Waveform "approximant" string understood by lalsimulation
            This is defines which waveform model is used.
        domain : Domain
            Domain object that specifies on which physical domain the
            waveform polarizations will be generated, e.g. Fourier
            domain, time domain.
        f_ref : float
            Reference frequency for the waveforms
        f_start : float
            Starting frequency for waveform generation. This is optional, and if not
            included, the starting frequency will be set to f_min. This exists so that
            EOB waveforms can be generated starting from a lower frequency than f_min.
        mode_list : List[Tuple]
            A list of waveform (ell, m) modes to include when generating
            the polarizations.
        spin_conversion_phase : float = None
            Value for phiRef when computing cartesian spins from bilby spins via
            bilby_to_lalsimulation_spins. The common convention is to use the value of
            the phase parameter here, which is also used in the spherical harmonics
            when combining the different modes. If spin_conversion_phase = None,
            this default behavior is adapted.
            For dingo, this convention for the phase parameter makes it impossible to
            treat the phase as an extrinsic parameter, since we can only account for
            the change of phase in the spherical harmonics when changing the phase (in
            order to also change the cartesian spins -- specifically, to rotate the spins
            by phase in the sx-sy plane -- one would need to recompute the modes,
            which is expensive).
            By setting spin_conversion_phase != None, we impose the convention to always
            use phase = spin_conversion_phase when computing the cartesian spins.
        """
        if not isinstance(approximant, str):
            raise ValueError("approximant should be a string, but got", approximant)
        else:
            self.approximant_str = approximant
            self.lal_params = None
            try:
                self.approximant = LS.GetApproximantFromString(approximant)
                if mode_list is not None:
                    self.lal_params = self.setup_mode_array(mode_list)
            except:
                pass

        if not issubclass(type(domain), Domain):
            raise ValueError(
                "domain should be an instance of a subclass of Domain, but got",
                type(domain),
            )
        else:
            self.domain = domain

        self.f_ref = f_ref
        self.f_start = f_start

        self.transform = transform
        self._spin_conversion_phase = None
        self.spin_conversion_phase = spin_conversion_phase

    @property
    def spin_conversion_phase(self):
        return self._spin_conversion_phase

    @spin_conversion_phase.setter
    def spin_conversion_phase(self, value):
        if value is None:
            print(
                "Setting spin_conversion_phase = None. Using phase parameter for "
                "conversion to cartesian spins."
            )
        else:
            print(
                f"Setting spin_conversion_phase = {value}. Using this value for the "
                f"phase parameter for conversion to cartesian spins."
            )
        self._spin_conversion_phase = value

    def generate_hplus_hcross(
        self, parameters: Dict[str, float], catch_waveform_errors=True
    ) -> Dict[str, np.ndarray]:
        """Generate GW polarizations (h_plus, h_cross).

        If the generation of the lalsimulation waveform fails with an
        "Input domain error", we return NaN polarizations.

        Use the domain, approximant, and mode_list specified in the constructor
        along with the waveform parameters to generate the waveform polarizations.


        Parameters
        ----------
        parameters: Dict[str, float]
            A dictionary of parameter names and scalar values.
            The parameter dictionary must include the following keys.
            For masses, spins, and distance there are multiple options.

            Mass: (mass_1, mass_2) or a pair of quantities from
                ((chirp_mass, total_mass), (mass_ratio, symmetric_mass_ratio))
            Spin:
                (a_1, a_2, tilt_1, tilt_2, phi_12, phi_jl) if precessing binary or
                (chi_1, chi_2) if the binary has aligned spins
            Reference frequency: f_ref at which spin vectors are defined
            Extrinsic:
                Distance: one of (luminosity_distance, redshift, comoving_distance)
                Inclination: theta_jn
                Reference phase: phase
                Geocentric time: geocent_time (GPS time)
            The following parameters are not required:
                Sky location: ra, dec,
                Polarization angle: psi
            Units:
                Masses should be given in units of solar masses.
                Distance should be given in megaparsecs (Mpc).
                Frequencies should be given in Hz and time in seconds.
                Spins should be dimensionless.
                Angles should be in radians.

        catch_waveform_errors: bool
            Whether to catch lalsimulation errors

        Returns
        -------
        wf_dict:
            A dictionary of generated waveform polarizations
        """
        if not isinstance(parameters, dict):
            raise ValueError("parameters should be a dictionary, but got", parameters)
        elif not isinstance(list(parameters.values())[0], float):
            raise ValueError("parameters dictionary must contain floats", parameters)

        # Include reference frequency with the parameters. Copy the dict first for safety.
        parameters = parameters.copy()
        parameters["f_ref"] = self.f_ref

        parameters_generator = self._convert_parameters(parameters, self.lal_params)

        # Generate GW polarizations
        if isinstance(self.domain, FrequencyDomain):
            wf_generator = self.generate_FD_waveform
        elif isinstance(self.domain, TimeDomain):
            wf_generator = self.generate_TD_waveform
        else:
            raise ValueError(f"Unsupported domain type {type(self.domain)}.")

        try:
            wf_dict = wf_generator(parameters_generator)
        except Exception as e:
            if not catch_waveform_errors:
                raise
            else:
                EDOM = e.args[0] == "Internal function call failed: Input domain error"
                if EDOM:
                    warnings.warn(
                        f"Evaluating the waveform failed with error: {e}\n"
                        f"The parameters were {parameters_generator}\n"
                    )
                    pol_nan = np.ones(len(self.domain)) * np.nan
                    wf_dict = {"h_plus": pol_nan, "h_cross": pol_nan}
                else:
                    raise

        if self.transform is not None:
            return self.transform(wf_dict)
        else:
            return wf_dict

    def _convert_to_scalar(self, x: Union[np.ndarray, float]) -> Number:
        """
        Convert a single element array to a number.

        Parameters
        ----------
        x:
            Array or number

        Returns
        -------
        A number
        """
        if isinstance(x, np.ndarray):
            if x.shape == () or x.shape == (1,):
                return x.item()
            else:
                raise ValueError(
                    f"Expected an array of length one, but shape = {x.shape}"
                )
        else:
            return x

    def _convert_parameters(
        self,
        parameter_dict: Dict,
        lal_params=None,
        lal_target_function=None,
    ) -> Tuple:
        """Convert to lal source frame parameters

        Parameters
        ----------
        parameter_dict : Dict
            A dictionary of parameter names and 1-dimensional prior distribution
            objects. If None, we use a default binary black hole prior.
        lal_params : (None, or Swig Object of type 'tagLALDict *')
            Extra parameters which can be passed to lalsimulation calls.
        lal_target_function: str = None
            Name of the lalsimulation function for which to prepare the parameters.
            If None, use SimInspiralFD if self.domain is FD, and SimInspiralTD if
            self.domain is TD.
            Choices:
                - SimInspiralFD (Also works for SimInspiralChooseFDWaveform)
                - SimInspiralTD (Also works for SimInspiralChooseTDWaveform)
                - SimInspiralChooseFDModes
                - SimInspiralChooseTDModes
        Returns
        -------
        lal_parameter_tuple:
            A tuple of parameters for the lalsimulation waveform generator
        """
        # check that the lal_target_function is valid
        if lal_target_function is None:
            if isinstance(self.domain, FrequencyDomain):
                lal_target_function = "SimInspiralFD"
            elif isinstance(self.domain, TimeDomain):
                lal_target_function = "SimInspiralTD"
            else:
                raise ValueError(f"Unsupported domain type {type(self.domain)}.")
        if lal_target_function not in [
            "SimInspiralFD",
            "SimInspiralTD",
            "SimInspiralChooseTDModes",
            "SimInspiralChooseFDModes",
            "SimIMRPhenomXPCalculateModelParametersFromSourceFrame",
        ]:
            raise ValueError(
                f"Unsupported lalsimulation waveform function {lal_target_function}."
            )

        # Transform mass, spin, and distance parameters
        p, _ = convert_to_lal_binary_black_hole_parameters(parameter_dict)

        # Convert to SI units
        p["mass_1"] *= lal.MSUN_SI
        p["mass_2"] *= lal.MSUN_SI
        p["luminosity_distance"] *= 1e6 * lal.PC_SI

        # Transform to lal source frame: iota and Cartesian spin components
        param_keys_in = (
            "theta_jn",
            "phi_jl",
            "tilt_1",
            "tilt_2",
            "phi_12",
            "a_1",
            "a_2",
            "mass_1",
            "mass_2",
            "f_ref",
            "phase",
        )
        param_values_in = [p[k] for k in param_keys_in]
        # if spin_conversion_phase is set, use this as fixed phiRef when computing the
        # cartesian spins instead of using the phase parameter
        if self.spin_conversion_phase is not None:
            param_values_in[-1] = self.spin_conversion_phase
        iota_and_cart_spins = bilby_to_lalsimulation_spins(*param_values_in)
        iota, s1x, s1y, s1z, s2x, s2y, s2z = [
            float(self._convert_to_scalar(x)) for x in iota_and_cart_spins
        ]

        # Construct argument list for FD and TD lal waveform generator wrappers
        spins_cartesian = s1x, s1y, s1z, s2x, s2y, s2z
        masses = (p["mass_1"], p["mass_2"])
        r = p["luminosity_distance"]
        phase = p["phase"]
        ecc_params = (0.0, 0.0, 0.0)  # longAscNodes, eccentricity, meanPerAno

        # Get domain parameters
        f_ref = p["f_ref"]
        if isinstance(self.domain, FrequencyDomain):
            delta_f = self.domain.delta_f
            f_max = self.domain.f_max
            if self.f_start is not None:
                f_min = self.f_start
            else:
                f_min = self.domain.f_min
            # parameters needed for TD waveforms
            delta_t = 0.5 / self.domain.f_max
        elif isinstance(self.domain, TimeDomain):
            raise NotImplementedError("Time domain not supported yet.")
            # FIXME: compute f_min from duration or specify it if SimInspiralTD
            #  is used for a native FD waveform
            f_min = 20.0
            delta_t = self.domain.delta_t
            # parameters needed for FD waveforms
            f_max = 1.0 / self.domain.delta_t
            delta_f = 1.0 / self.domain.duration
        else:
            raise ValueError(f"Unsupported domain type {type(self.domain)}.")

        if lal_target_function == "SimInspiralFD":
            # LS.SimInspiralFD takes parameters:
            #   m1, m2, S1x, S1y, S1z, S2x, S2y, S2z,
            #   distance, inclination, phiRef,
            #   longAscNodes, eccentricity, meanPerAno,
            #   deltaF, f_min, f_max, f_ref,
            #   lal_params, approximant
            domain_pars = (delta_f, f_min, f_max, f_ref)
            domain_pars = tuple(float(p) for p in domain_pars)
            lal_parameter_tuple = (
                masses
                + spins_cartesian
                + (r, iota, phase)
                + ecc_params
                + domain_pars
                + (lal_params, self.approximant)
            )

        elif lal_target_function == "SimInspiralTD":
            # LS.SimInspiralTD takes parameters:
            #   m1, m2, S1x, S1y, S1z, S2x, S2y, S2z,
            #   distance, inclination, phiRef,
            #   longAscNodes, eccentricity, meanPerAno,
            #   delta_t, f_min, f_ref
            #   lal_params, approximant
            domain_pars = (delta_t, f_min, f_ref)
            domain_pars = tuple(float(p) for p in domain_pars)
            lal_parameter_tuple = (
                masses
                + spins_cartesian
                + (r, iota, phase)
                + ecc_params
                + domain_pars
                + (lal_params, self.approximant)
            )
        elif lal_target_function == "SimInspiralChooseFDModes":
            domain_pars = (delta_f, f_min, f_max, f_ref)
            domain_pars = tuple(float(p) for p in domain_pars)
            lal_parameter_tuple = (
                masses
                + spins_cartesian
                + domain_pars
                + (phase, r, iota)
                + (lal_params, self.approximant)
            )

        elif (
            lal_target_function
            == "SimIMRPhenomXPCalculateModelParametersFromSourceFrame"
        ):
            lal_parameter_tuple = (
                masses + (f_ref,) + (phase, iota) + spins_cartesian + (lal_params,)
            )

        elif lal_target_function == "SimInspiralChooseTDModes":
            # LS.SimInspiralChooseTDModes takes parameters:
            #   phiRef=0 (for lal legacy reasons), delta_t,
            #   m1, m2, S1x, S1y, S1z, S2x, S2y, S2z,
            #   f_min, f_ref
            #   distance,
            #   lal_params, l_max, approximant
            domain_pars = (delta_t, f_min, f_ref)
            domain_pars = tuple(float(p) for p in domain_pars)
            if "l_max" not in parameter_dict:
                l_max = 5  # hard code l_max for now
            lal_parameter_tuple = (
                (
                    0.0,
                    domain_pars[0],
                )  # domain_pars[0] = delta_t
                + masses
                + spins_cartesian
                + domain_pars[1:]  # domain_pars[1:] = f_min, f_ref
                + (r,)
                + (lal_params, l_max, self.approximant)
            )
            # also pass iota, since this is needed for recombination of the modes
            lal_parameter_tuple = (lal_parameter_tuple, iota)

        return lal_parameter_tuple

    def setup_mode_array(self, mode_list: List[Tuple]) -> lal.Dict:
        """Define a mode array to select waveform modes
        to include in the polarizations from a list of modes.

        Parameters
        ----------
        mode_list : a list of (ell, m) modes

        Returns
        -------
        lal_params:
            A lal parameter dictionary
        """
        lal_params = lal.CreateDict()
        ma = LS.SimInspiralCreateModeArray()
        for ell, m in mode_list:
            LS.SimInspiralModeArrayActivateMode(ma, ell, m)
            # LS.SimInspiralModeArrayActivateMode(ma, ell, -m)
        LS.SimInspiralWaveformParamsInsertModeArray(lal_params, ma)
        return lal_params

    def generate_FD_waveform(self, parameters_lal: Tuple) -> Dict[str, np.ndarray]:
        """
        Generate Fourier domain GW polarizations (h_plus, h_cross).

        Parameters
        ----------
        parameters_lal:
            A tuple of parameters for the lalsimulation waveform generator

        Returns
        -------
        pol_dict:
            A dictionary of generated waveform polarizations
        """
        # Note: SEOBNRv4PHM does not support the specification of spins at a
        # reference frequency different from the starting frequency. In addition,
        # waveform generation will fail if the orbital distance is smaller than ~ 10M.
        # To avoid this, we can start at a sufficiently low and consistent starting frequency
        # for the entire dataset. If the number of generation failures is a very small
        # fraction over the prior distribution then the dataset should be good to use.
        #
        # Note: XLALSimInspiralFD() internally calls XLALSimInspiralTD() to generate
        # a conditioned time-domain waveform. In the past, this function lowered
        # the starting frequency, but this is thankfully no longer the case
        # for models such as SEOBNRv4PHM where the reference frequency is equal
        # to the starting frequency. So, the TD waveform will be generated by
        # calling XLALSimInspiralChooseTDWaveform().
        # See https://git.ligo.org/waveforms/reviews/lalsuite/-/commit/195f9127682de19f5fce19cc5828116dd2d23461
        #
        # LS.SimInspiralFD takes parameters:
        #   m1, m2, S1x, S1y, S1z, S2x, S2y, S2z,
        #   distance, inclination, phiRef,
        #   longAscNodes, eccentricity, meanPerAno,
        #   deltaF, f_min, f_max, f_ref,
        #   lal_params, approximant

        # Sanity check types of arguments
        check_floats = all(map(lambda x: isinstance(x, float), parameters_lal[:18]))
        check_int = isinstance(parameters_lal[19], int)
        # parameters_lal[18]  # lal_params could be None or a LALDict
        if not (check_floats and check_int):
            raise ValueError(
                "SimInspiralFD received invalid argument(s)", parameters_lal
            )

        # Depending on whether the domain is uniform or non-uniform call the appropriate wf generator
        hp, hc = LS.SimInspiralFD(*parameters_lal)
        # The check below filters for unphysical waveforms:
        # For IMRPhenomXPHM, the LS.SimInspiralFD result is numerically instable
        # for rare parameter configurations (~1 in 1M), leading to bins with very large
        # numbers if multibanding is used. If that happens, turn off multibanding to
        # fix this.
        if max(np.max(np.abs(hp.data.data)), np.max(np.abs(hc.data.data))) > 1e-20:
            print(
                f"Generation with parameters {parameters_lal} likely numerically "
                f"unstable due to multibanding, turn off multibanding."
            )
            lal_dict = parameters_lal[-2]
            if lal_dict is None:
                lal_dict = lal.CreateDict()
            LS.SimInspiralWaveformParamsInsertPhenomXHMThresholdMband(lal_dict, 0)
            LS.SimInspiralWaveformParamsInsertPhenomXPHMThresholdMband(lal_dict, 0)
            hp, hc = LS.SimInspiralFD(
                *parameters_lal[:-2], lal_dict, parameters_lal[-1]
            )
            if max(np.max(np.abs(hp.data.data)), np.max(np.abs(hc.data.data))) > 1e-20:
                print(
                    f"Warning: turning off multibanding for parameters {parameters_lal}"
                    f"likely numerically might not have fixed it, check manually."
                )

        # Ensure that the waveform agrees with the frequency grid defined in the domain.
        if not isclose(self.domain.delta_f, hp.deltaF, rel_tol=1e-6):
            raise ValueError(
                f"Waveform delta_f is inconsistent with domain: {hp.deltaF} vs {self.domain.delta_f}!"
                f"To avoid this, ensure that f_max = {self.domain.f_max} is a power of two"
                "when you are using a native time-domain waveform model."
            )

        frequency_array = self.domain()
        h_plus = np.zeros_like(frequency_array, dtype=complex)
        h_cross = np.zeros_like(frequency_array, dtype=complex)
        # Ensure that length of wf agrees with length of domain. Enforce by truncating frequencies beyond f_max
        if len(hp.data.data) > len(frequency_array):
            warnings.warn(
                "LALsimulation waveform longer than domain's `frequency_array`"
                f"({len(hp.data.data)} vs {len(frequency_array)}). Truncating lalsim array."
            )
            h_plus = hp.data.data[: len(h_plus)]
            h_cross = hc.data.data[: len(h_cross)]
        else:
            h_plus[: len(hp.data.data)] = hp.data.data
            h_cross[: len(hc.data.data)] = hc.data.data

        # Undo the time shift done in SimInspiralFD to the waveform
        dt = 1 / hp.deltaF + (hp.epoch.gpsSeconds + hp.epoch.gpsNanoSeconds * 1e-9)
        time_shift = np.exp(-1j * 2 * np.pi * dt * frequency_array)
        h_plus *= time_shift
        h_cross *= time_shift
        pol_dict = {"h_plus": h_plus, "h_cross": h_cross}
        return pol_dict

    def generate_hplus_hcross_m(
        self, parameters: Dict[str, float]
    ) -> Dict[tuple, Dict[str, np.ndarray]]:
        """
        Generate GW polarizations (h_plus, h_cross), separated into contributions from
        the different modes. This method is identical to self.generate_hplus_hcross,
        except that it generates the individual contributions of the modes to the
        polarizations and sorts these according to their transformation behavior (see
        below), instead of returning the overall sum.

        This is useful in order to treat the phase as an extrinsic parameter. Instead of
        {"h_plus": hp, "h_cross": hc}, this method returns a dict in the form of
        {m: {"h_plus": hp_m, "h_cross": hc_m} for m in [-l_max,...,0,...,l_max]}. Each
        key m contains the contribution to the polarization that transforms according
        to exp(-1j * m * phase) under phase transformations (due to the spherical
        harmonics).

        Note:
            - pol_m[m] contains contributions of the m modes *and* and the -m modes.
              This is because the frequency domain (FD) modes have a positive frequency
              part which transforms as exp(-1j * m * phase), while the negative
              frequency part transforms as exp(+1j * m * phase). Typically, one of these
              dominates [e.g., the (2,2) mode is dominated by the negative frequency
              part and the (-2,2) mode is dominated by the positive frequency part]
              such that the sum of (l,|m|) and (l,-|m|) modes transforms approximately as
              exp(1j * |m| * phase), which is e.g. used for phase marginalization in
              bilby/lalinference. However, this is not exact. In this method we account
              for this effect, such that each contribution pol_m[m] transforms
              *exactly* as exp(-1j * m * phase).
            - Phase shifts contribute in two ways: Firstly via the spherical harmonics,
              which we account for with the exp(-1j * m * phase) transformation.
              Secondly, the phase determines how the PE spins transform to cartesian
              spins, by rotating (sx,sy) by phase. This is *not* accounted for in this
              function. Instead, the phase for computing the cartesian spins is fixed
              to self.spin_conversion_phase (if not None). This effectively changes the
              PE parameters {phi_jl, phi_12} to parameters {phi_jl_prime, phi_12_prime}.
              For parameter estimation, a postprocessing operation can be applied to
              account for this, {phi_jl_prime, phi_12_prime} -> {phi_jl, phi_12}.
              See also documentation of __init__ method for more information on
              self.spin_conversion_phase.

        Differences to self.generate_hplus_hcross:
        - We don't catch errors yet TODO
        - We don't apply transforms yet TODO

        Parameters
        ----------
        parameters: dict
            Dictionary of parameters for the waveform.
            For details see see self.generate_hplus_hcross.

        Returns
        -------
        pol_m: dict
            Dictionary with contributions to h_plus and h_cross, sorted by their
            transformation behaviour under phase shifts:
            {m: {"h_plus": hp_m, "h_cross": hc_m} for m in [-l_max,...,0,...,l_max]}
            Each contribution h_m transforms as exp(-1j * m * phase) under phase shifts
            (for fixed self.spin_conversion_phase, see above).
        """
        if not isinstance(parameters, dict):
            raise ValueError("parameters should be a dictionary, but got", parameters)
        elif not isinstance(list(parameters.values())[0], float):
            raise ValueError("parameters dictionary must contain floats", parameters)

        if isinstance(self.domain, FrequencyDomain):
            # Generate FD modes in for frequencies [-f_max, ..., 0, ..., f_max].
            if LS.SimInspiralImplementedFDApproximants(self.approximant):
                # Step 1: generate waveform modes in L0 frame in native domain of
                # approximant (here: FD)
                hlm_fd, iota = self.generate_FD_modes_LO(parameters)

                # Step 2: Transform modes to target domain.
                # Not required here, as approximant domain and target domain are both FD.

            else:
                assert LS.SimInspiralImplementedTDApproximants(self.approximant)
                # Step 1: generate waveform modes in L0 frame in native domain of
                # approximant (here: TD)
                hlm_td, iota = self.generate_TD_modes_L0(parameters)

                # Step 2: Transform modes to target domain.
                # This requires tapering of TD modes, and FFT to transform to FD.
                wfg_utils.taper_td_modes_in_place(hlm_td)
                hlm_fd = wfg_utils.td_modes_to_fd_modes(hlm_td, self.domain)

            # Step 3: Separate negative and positive frequency parts of the modes,
            # and add contributions according to their transformation behavior under
            # phase shifts.
            pol_m = wfg_utils.get_polarizations_from_fd_modes_m(
                hlm_fd, iota, parameters["phase"]
            )

        else:
            raise NotImplementedError(
                f"Target domain of type {type(self.domain)} not yet implemented."
            )

        return pol_m

    def generate_FD_modes_LO(self, parameters):
        """
        Generate FD modes in the L0 frame.

        Parameters
        ----------
        parameters: dict
            Dictionary of parameters for the waveform.
            For details see see self.generate_hplus_hcross.

        Returns
        -------
        hlm_fd: dict
            Dictionary with (l,m) as keys and the corresponding FD modes in lal format as
            values.
        iota: float
        """
        # TD approximants that are implemented in J frame. Currently tested for:
        #   101: IMRPhenomXPHM
        if self.approximant in [101]:
            parameters_lal_fd_modes = self._convert_parameters(
                {**parameters, "f_ref": self.f_ref},
                lal_target_function="SimInspiralChooseFDModes",
            )
            iota = parameters_lal_fd_modes[14]
            hlm_fd = LS.SimInspiralChooseFDModes(*parameters_lal_fd_modes)
            # unpack linked list, convert lal objects to arrays
            hlm_fd = wfg_utils.linked_list_modes_to_dict_modes(hlm_fd)
            hlm_fd = {k: v.data.data for k, v in hlm_fd.items()}
            # For the waveform models considered here (e.g., IMRPhenomXPHM), the modes
            # are returned in the J frame (where the observer is at inclination=theta_JN,
            # azimuth=0). In this frame, the dependence on the reference phase enters
            # via the modes themselves. We need to convert to the L0 frame so that the
            # dependence on phase enters via the spherical harmonics.
            hlm_fd = frame_utils.convert_J_to_L0_frame(
                hlm_fd,
                parameters,
                self,
                spin_conversion_phase=self.spin_conversion_phase,
            )
            return hlm_fd, iota
        else:
            raise NotImplementedError(
                f"Approximant {LS.GetApproximantFromString(self.approximant)} not "
                f"implemented. When adding this approximant to this method, make sure "
                f"the the output dict hlm_td contains the TD modes in the *L0 frame*. "
                f"In particular, adding an approximant that is implemented in the same "
                f"domain and frame as one of the approximants should just be a matter of "
                f"adding the approximant number (here: {self.approximant}) to the "
                f"corresponding if statement. However, when doing this please make sure "
                f"to test that this works as intended! Ideally, add some unit tests."
            )

    def generate_TD_modes_L0(self, parameters):
        """
        Generate TD modes in the L0 frame.

        Parameters
        ----------
        parameters: dict
            Dictionary of parameters for the waveform.
            For details see see self.generate_hplus_hcross.

        Returns
        -------
        hlm_td: dict
            Dictionary with (l,m) as keys and the corresponding TD modes in lal format as
            values.
        iota: float
        """
        # TD approximants that are implemented in L0 frame. Currently tested for:
        #   52: SEOBNRv4PHM
        if self.approximant in [52]:
            parameters_lal_td_modes, iota = self._convert_parameters(
                {**parameters, "f_ref": self.f_ref},
                lal_target_function="SimInspiralChooseTDModes",
            )
            hlm_td = LS.SimInspiralChooseTDModes(*parameters_lal_td_modes)
            return wfg_utils.linked_list_modes_to_dict_modes(hlm_td), iota
        else:
            raise NotImplementedError(
                f"Approximant {LS.GetApproximantFromString(self.approximant)} not "
                f"implemented. When adding this approximant to this method, make sure "
                f"the the output dict hlm_td contains the TD modes in the *L0 frame*. "
                f"In particular, adding an approximant that is implemented in the same "
                f"domain and frame as one of the approximants should just be a matter of "
                f"adding the approximant number (here: {self.approximant}) to the "
                f"corresponding if statement. However, when doing this please make sure "
                f"to test that this works as intended! Ideally, add some unit tests."
            )

    def generate_TD_waveform(self, parameters_lal: Tuple) -> Dict[str, np.ndarray]:
        """
        Generate time domain GW polarizations (h_plus, h_cross)

        Parameters
        ----------
        parameters_lal:
            A tuple of parameters for the lalsimulation waveform generator

        Returns
        -------
        pol_dict:
            A dictionary of generated waveform polarizations
        """
        # Note: XLALSimInspiralTD() now calls XLALSimInspiralChooseTDWaveform()
        # for models such as SEOBNRv4PHM where the reference frequency is equal
        # to the starting frequency and thus leaves our choice of starting
        # frequency untouched.
        #
        # LS.SimInspiralTD takes parameters:
        #   m1, m2, S1x, S1y, S1z, S2x, S2y, S2z,
        #   distance, inclination, phiRef,
        #   longAscNodes, eccentricity, meanPerAno,
        #   deltaT, f_min, f_ref
        #   lal_params, approximant

        hp, hc = LS.SimInspiralTD(*parameters_lal)
        h_plus = hp.data.data
        h_cross = hc.data.data
        pol_dict = {"h_plus": h_plus, "h_cross": h_cross}
        return pol_dict


class NewInterfaceWaveformGenerator(WaveformGenerator):
    """Generate polarizations using GWSignal routines in the specified domain for a
    single GW coalescence given a set of waveform parameters.
    """

    def __init__(self, **kwargs):
        WaveformGenerator.__init__(self, **kwargs)

        self.mode_list = kwargs.get("mode_list", None)

    def _convert_parameters(
        self,
        parameter_dict: Dict,
        lal_params=None,
    ):
        # Transform mass, spin, and distance parameters
        p, _ = convert_to_lal_binary_black_hole_parameters(parameter_dict)
        # Transform to lal source frame: iota and Cartesian spin components
        param_keys_in = (
            "theta_jn",
            "phi_jl",
            "tilt_1",
            "tilt_2",
            "phi_12",
            "a_1",
            "a_2",
            "mass_1",
            "mass_2",
            "f_ref",
            "phase",
        )
        param_values_in = [p[k] for k in param_keys_in]
        # if spin_conversion_phase is set, use this as fixed phiRef when computing the
        # cartesian spins instead of using the phase parameter
        if self.spin_conversion_phase is not None:
            param_values_in[-1] = self.spin_conversion_phase
        iota_and_cart_spins = bilby_to_lalsimulation_spins(*param_values_in)
        iota, s1x, s1y, s1z, s2x, s2y, s2z = [
            float(self._convert_to_scalar(x)) for x in iota_and_cart_spins
        ]

        f_ref = p["f_ref"]
        delta_f = self.domain.delta_f
        f_max = self.domain.f_max
        if self.f_start is not None:
            f_min = self.f_start
        else:
            f_min = self.domain.f_min
        # parameters needed for TD waveforms
        delta_t = 0.5 / self.domain.f_max

        params_gwsignal = {
            "mass1": p["mass_1"] * u.solMass,
            "mass2": p["mass_2"] * u.solMass,
            "spin1x": s1x * u.dimensionless_unscaled,
            "spin1y": s1y * u.dimensionless_unscaled,
            "spin1z": s1z * u.dimensionless_unscaled,
            "spin2x": s2x * u.dimensionless_unscaled,
            "spin2y": s2y * u.dimensionless_unscaled,
            "spin2z": s2z * u.dimensionless_unscaled,
            "deltaT": delta_t * u.s,
            "f22_start": f_min * u.Hz,
            "f22_ref": f_ref * u.Hz,
            "f_max": f_max * u.Hz,
            "deltaF": delta_f * u.Hz,
            "phi_ref": p["phase"] * u.rad,
            "distance": p["luminosity_distance"] * u.Mpc,
            "inclination": iota * u.rad,
            "ModeArray": self.mode_list,
            "condition": 1,
        }

        # SEOBNRv5 specific parameters
        if "postadiabatic" in p:
            params_gwsignal["postadiabatic"] = p["postadiabatic"]

            if "postadiabatic_type" in p:
                params_gwsignal["postadiabatic_type"] = p["postadiabatic_type"]

        if "lmax_nyquist" in p:
            params_gwsignal["lmax_nyquist"] = p["lmax_nyquist"]
        else:
            params_gwsignal["lmax_nyquist"] = 2

        return params_gwsignal

    def generate_FD_waveform(self, parameters_gwsignal: Dict) -> Dict[str, np.ndarray]:
        """
        Generate Fourier domain GW polarizations (h_plus, h_cross).

        Parameters
        ----------
        parameters_lal:
            A tuple of parameters for the lalsimulation waveform generator

        Returns
        -------
        pol_dict:
            A dictionary of generated waveform polarizations
        """
        # Note: SEOBNRv4PHM does not support the specification of spins at a
        # reference frequency different from the starting frequency. In addition,
        # waveform generation will fail if the orbital distance is smaller than ~ 10M.
        # To avoid this, we can start at a sufficiently low and consistent starting frequency
        # for the entire dataset. If the number of generation failures is a very small
        # fraction over the prior distribution then the dataset should be good to use.
        #
        # Note: XLALSimInspiralFD() internally calls XLALSimInspiralTD() to generate
        # a conditioned time-domain waveform. In the past, this function lowered
        # the starting frequency, but this is thankfully no longer the case
        # for models such as SEOBNRv4PHM where the reference frequency is equal
        # to the starting frequency. So, the TD waveform will be generated by
        # calling XLALSimInspiralChooseTDWaveform().
        # See https://git.ligo.org/waveforms/reviews/lalsuite/-/commit/195f9127682de19f5fce19cc5828116dd2d23461
        #
        # LS.SimInspiralFD takes parameters:
        #   m1, m2, S1x, S1y, S1z, S2x, S2y, S2z,
        #   distance, inclination, phiRef,
        #   longAscNodes, eccentricity, meanPerAno,
        #   deltaF, f_min, f_max, f_ref,
        #   lal_params, approximant

        # Sanity check types of arguments
        # check_floats = all(map(lambda x: isinstance(x, float), parameters_lal[:18]))
        # check_int = isinstance(parameters_lal[19], int)
        # parameters_lal[18]  # lal_params could be None or a LALDict
        # if not (check_floats and check_int):
        #    raise ValueError(
        #        "SimInspiralFD received invalid argument(s)", parameters_lal
        #    )

        # Depending on whether the domain is uniform or non-uniform call the appropriate wf generator
        generator = new_interface_get_waveform_generator(self.approximant_str)
        hpc = gws_wfm.GenerateFDWaveform(parameters_gwsignal, generator)
        hp = hpc.hp
        hc = hpc.hc

        # Ensure that the waveform agrees with the frequency grid defined in the domain.
        if not isclose(self.domain.delta_f, hp.df.value, rel_tol=1e-6):
            raise ValueError(
                f"Waveform delta_f is inconsistent with domain: {hp.df.value} vs {self.domain.delta_f}!"
                f"To avoid this, ensure that f_max = {self.domain.f_max} is a power of two"
                "when you are using a native time-domain waveform model."
            )

        frequency_array = self.domain()
        h_plus = np.zeros_like(frequency_array, dtype=complex)
        h_cross = np.zeros_like(frequency_array, dtype=complex)
        # Ensure that length of wf agrees with length of domain. Enforce by truncating frequencies beyond f_max
        if len(hp) > len(frequency_array):
            warnings.warn(
                "GWSignal waveform longer than domain's `frequency_array`"
                f"({len(hp)} vs {len(frequency_array)}). Truncating gwsignal array."
            )
            h_plus = hp[: len(h_plus)].value
            h_cross = hc[: len(h_cross)].value
        else:
            h_plus = hp.value
            h_cross = hc.value

        # Undo the time shift done in SimInspiralFD to the waveform
        dt = 1 / hp.df.value + hp.epoch.value
        time_shift = np.exp(-1j * 2 * np.pi * dt * frequency_array)
        h_plus *= time_shift
        h_cross *= time_shift
        pol_dict = {"h_plus": h_plus, "h_cross": h_cross}
        return pol_dict

    def generate_hplus_hcross_m(
        self, parameters: Dict[str, float]
    ) -> Dict[tuple, Dict[str, np.ndarray]]:
        """
        Generate GW polarizations (h_plus, h_cross), separated into contributions from
        the different modes. This method is identical to self.generate_hplus_hcross,
        except that it generates the individual contributions of the modes to the
        polarizations and sorts these according to their transformation behavior (see
        below), instead of returning the overall sum.

        This is useful in order to treat the phase as an extrinsic parameter. Instead of
        {"h_plus": hp, "h_cross": hc}, this method returns a dict in the form of
        {m: {"h_plus": hp_m, "h_cross": hc_m} for m in [-l_max,...,0,...,l_max]}. Each
        key m contains the contribution to the polarization that transforms according
        to exp(-1j * m * phase) under phase transformations (due to the spherical
        harmonics).

        Note:
            - pol_m[m] contains contributions of the m modes *and* and the -m modes.
              This is because the frequency domain (FD) modes have a positive frequency
              part which transforms as exp(-1j * m * phase), while the negative
              frequency part transforms as exp(+1j * m * phase). Typically, one of these
              dominates [e.g., the (2,2) mode is dominated by the negative frequency
              part and the (-2,2) mode is dominated by the positive frequency part]
              such that the sum of (l,|m|) and (l,-|m|) modes transforms approximately as
              exp(1j * |m| * phase), which is e.g. used for phase marginalization in
              bilby/lalinference. However, this is not exact. In this method we account
              for this effect, such that each contribution pol_m[m] transforms
              *exactly* as exp(-1j * m * phase).
            - Phase shifts contribute in two ways: Firstly via the spherical harmonics,
              which we account for with the exp(-1j * m * phase) transformation.
              Secondly, the phase determines how the PE spins transform to cartesian
              spins, by rotating (sx,sy) by phase. This is *not* accounted for in this
              function. Instead, the phase for computing the cartesian spins is fixed
              to self.spin_conversion_phase (if not None). This effectively changes the
              PE parameters {phi_jl, phi_12} to parameters {phi_jl_prime, phi_12_prime}.
              For parameter estimation, a postprocessing operation can be applied to
              account for this, {phi_jl_prime, phi_12_prime} -> {phi_jl, phi_12}.
              See also documentation of __init__ method for more information on
              self.spin_conversion_phase.

        Differences to self.generate_hplus_hcross:
        - We don't catch errors yet TODO
        - We don't apply transforms yet TODO

        Parameters
        ----------
        parameters: dict
            Dictionary of parameters for the waveform.
            For details see see self.generate_hplus_hcross.

        Returns
        -------
        pol_m: dict
            Dictionary with contributions to h_plus and h_cross, sorted by their
            transformation behaviour under phase shifts:
            {m: {"h_plus": hp_m, "h_cross": hc_m} for m in [-l_max,...,0,...,l_max]}
            Each contribution h_m transforms as exp(-1j * m * phase) under phase shifts
            (for fixed self.spin_conversion_phase, see above).
        """
        if not isinstance(parameters, dict):
            raise ValueError("parameters should be a dictionary, but got", parameters)
        elif not isinstance(list(parameters.values())[0], float):
            raise ValueError("parameters dictionary must contain floats", parameters)

        generator = new_interface_get_waveform_generator(self.approximant_str)
        if isinstance(self.domain, FrequencyDomain):
            # Generate FD modes in for frequencies [-f_max, ..., 0, ..., f_max].
            if generator.domain == "freq":
                # Step 1: generate waveform modes in L0 frame in native domain of
                # approximant (here: FD)
                hlm_fd, iota = self.generate_FD_modes_LO(parameters)

                # Step 2: Transform modes to target domain.
                # Not required here, as approximant domain and target domain are both FD.

            else:
                # assert LS.SimInspiralImplementedTDApproximants(self.approximant)
                # Step 1: generate waveform modes in L0 frame in native domain of
                # approximant (here: TD)
                hlm_td, iota = self.generate_TD_modes_L0(parameters)

                # Step 2: Transform modes to target domain.
                # This requires tapering of TD modes, and FFT to transform to FD.
                wfg_utils.taper_td_modes_in_place(hlm_td)
                hlm_fd = wfg_utils.td_modes_to_fd_modes(hlm_td, self.domain)

            # Step 3: Separate negative and positive frequency parts of the modes,
            # and add contributions according to their transformation behavior under
            # phase shifts.
            pol_m = wfg_utils.get_polarizations_from_fd_modes_m(
                hlm_fd, iota, parameters["phase"]
            )

        else:
            raise NotImplementedError(
                f"Target domain of type {type(self.domain)} not yet implemented."
            )

        return pol_m

    def generate_FD_modes_LO(self, parameters):  # Pending to adapt
        """
        Generate FD modes in the L0 frame.

        Parameters
        ----------
        parameters: dict
            Dictionary of parameters for the waveform.
            For details see see self.generate_hplus_hcross.

        Returns
        -------
        hlm_fd: dict
            Dictionary with (l,m) as keys and the corresponding FD modes in lal format as
            values.
        iota: float
        """
        # TD approximants that are implemented in J frame. Currently tested for:
        #   101: IMRPhenomXPHM
        if self.approximant_str in ["IMRPhenomXPHM"]:
            parameters_gwsignal = self._convert_parameters(
                {**parameters, "f_ref": self.f_ref}
            )
            iota = parameters_gwsignal["inclination"]
            generator = new_interface_get_waveform_generator(self.approximant_str)
            hlm_fd = gws_wfm.GenerateFDModes(parameters_gwsignal, generator)
            # unpack linked list, convert lal objects to arrays

            hlms_lal = {}
            for key, value in hlm_fd.items():
                if type(key) != str:
                    hlm_lal = lal.CreateCOMPLEX16TimeSeries(
                        "hplus",
                        value.epoch.value,
                        0,
                        value.dt.value,
                        lal.DimensionlessUnit,
                        len(value),
                    )
                    hlm_lal.data.data = value.value
                    hlms_lal[key] = hlm_lal

            hlm_fd = wfg_utils.linked_list_modes_to_dict_modes(hlms_lal)
            hlm_fd = {k: v.data.data for k, v in hlm_fd.items()}
            # For the waveform models considered here (e.g., IMRPhenomXPHM), the modes
            # are returned in the J frame (where the observer is at inclination=theta_JN,
            # azimuth=0). In this frame, the dependence on the reference phase enters
            # via the modes themselves. We need to convert to the L0 frame so that the
            # dependence on phase enters via the spherical harmonics.
            hlm_fd = frame_utils.convert_J_to_L0_frame(
                hlm_fd,
                parameters,
                self,
                spin_conversion_phase=self.spin_conversion_phase,
            )
            return hlm_fd, iota
        else:
            raise NotImplementedError(
                f"Approximant {LS.GetApproximantFromString(self.approximant)} not "
                f"implemented. When adding this approximant to this method, make sure "
                f"the the output dict hlm_td contains the TD modes in the *L0 frame*. "
                f"In particular, adding an approximant that is implemented in the same "
                f"domain and frame as one of the approximants should just be a matter of "
                f"adding the approximant number (here: {self.approximant}) to the "
                f"corresponding if statement. However, when doing this please make sure "
                f"to test that this works as intended! Ideally, add some unit tests."
            )

    def generate_TD_modes_L0(self, parameters):
        """
        Generate TD modes in the L0 frame.

        Parameters
        ----------
        parameters: dict
            Dictionary of parameters for the waveform.
            For details see see self.generate_hplus_hcross.

        Returns
        -------
        hlm_td: dict
            Dictionary with (l,m) as keys and the corresponding TD modes in lal format as
            values.
        iota: float
        """
        # TD approximants that are implemented in L0 frame. Currently tested for:
        #   52: SEOBNRv4PHM

        parameters_gwsignal = self._convert_parameters(
            {**parameters, "f_ref": self.f_ref}
        )

        generator = new_interface_get_waveform_generator(self.approximant_str)
        hlm_td = gws_wfm.GenerateTDModes(parameters_gwsignal, generator)
        hlms_lal = {}

        for key, value in hlm_td.items():
            if type(key) != str:
                hlm_lal = lal.CreateCOMPLEX16TimeSeries(
                    "hplus",
                    value.epoch.value,
                    0,
                    value.dt.value,
                    lal.DimensionlessUnit,
                    len(value),
                )
                hlm_lal.data.data = value.value
                hlms_lal[key] = hlm_lal

        return hlms_lal, parameters_gwsignal["inclination"].value

    def generate_TD_waveform(self, parameters_gwsignal: Dict) -> Dict[str, np.ndarray]:
        """
        Generate time domain GW polarizations (h_plus, h_cross)

        Parameters
        ----------
        parameters_gwsignal:
            A dict of parameters for the gwsignal waveform generator

        Returns
        -------
        pol_dict:
            A dictionary of generated waveform polarizations
        """
        # Note: XLALSimInspiralTD() now calls XLALSimInspiralChooseTDWaveform()
        # for models such as SEOBNRv4PHM where the reference frequency is equal
        # to the starting frequency and thus leaves our choice of starting
        # frequency untouched.
        #
        # LS.SimInspiralTD takes parameters:
        #   m1, m2, S1x, S1y, S1z, S2x, S2y, S2z,
        #   distance, inclination, phiRef,
        #   longAscNodes, eccentricity, meanPerAno,
        #   deltaT, f_min, f_ref
        #   lal_params, approximant

        generator = new_interface_get_waveform_generator(self.approximant_str)
        hpc = gws_wfm.GenerateTDWaveform(parameters_gwsignal, generator)

        h_plus = hpc.hp.value
        h_cross = hpc.hc.value
        pol_dict = {"h_plus": h_plus, "h_cross": h_cross}
        return pol_dict


def SEOBNRv4PHM_maximum_starting_frequency(
    total_mass: float, fudge: float = 0.99
) -> float:
    """
    Given a total mass return the largest possible starting frequency allowed
    for SEOBNRv4PHM and similar effective-one-body models.

    The intended use for this function is at the stage of designing
    a data set: after choosing a mass prior one can use it to figure out
    which prior samples would run into an issue when generating an EOB waveform,
    and tweak the parameters to reduce the number of failing configurations.

    Parameters
    ----------
    total_mass:
        Total mass in units of solar masses
    fudge:
        A fudge factor

    Returns
    -------
    f_max_Hz:
        The largest possible starting frequency in Hz
    """
    total_mass_sec = total_mass * lal.MTSUN_SI
    f_max_Hz = fudge * 10.5 ** (-1.5) / (np.pi * total_mass_sec)
    return f_max_Hz


def generate_waveforms_task_func(
    args: Tuple, waveform_generator: WaveformGenerator
) -> Dict[str, np.ndarray]:
    """
    Picklable wrapper function for parallel waveform generation.

    Parameters
    ----------
    args:
        A tuple of (index, pandas.core.series.Series)
    waveform_generator:
        A WaveformGenerator instance

    Returns
    -------
    The generated waveform polarization dictionary
    """
    parameters = args[1].to_dict()
    return waveform_generator.generate_hplus_hcross(parameters)


def generate_waveforms_parallel(
    waveform_generator: WaveformGenerator,
    parameter_samples: pd.DataFrame,
    pool: Pool = None,
) -> Dict[str, np.ndarray]:
    """Generate a waveform dataset, optionally in parallel.

    Parameters
    ----------
    waveform_generator: WaveformGenerator
        A WaveformGenerator instance
    parameter_samples: pd.DataFrame
        Intrinsic parameter samples
    pool: multiprocessing.Pool
        Optional pool of workers for parallel generation

    Returns
    -------
    polarizations:
        A dictionary of all generated polarizations stacked together
    """
    # logger.info('Generating waveform polarizations ...')

    task_func = partial(
        generate_waveforms_task_func, waveform_generator=waveform_generator
    )
    task_data = parameter_samples.iterrows()

    if pool is not None:
        polarizations_list = pool.map(task_func, task_data)
    else:
        polarizations_list = list(map(task_func, task_data))
    polarizations = {
        pol: np.stack([wf[pol] for wf in polarizations_list])
        for pol in polarizations_list[0].keys()
    }
    return polarizations


def sum_contributions_m(x_m, phase_shift=0.0):
    """
    Sum the contributions over m-components, optionally introducing a phase shift.
    """
    keys = next(iter(x_m.values())).keys()
    result = {key: 0.0 for key in keys}
    for key in keys:
        for m, x in x_m.items():
            result[key] += x[key] * np.exp(-1j * m * phase_shift)
    return result


if __name__ == "__main__":
    import pandas as pd
    import numpy as np
    from dingo.gw.domains import build_domain
    from dingo.gw.prior import build_prior_with_defaults

    domain_settings = {
        "type": "FrequencyDomain",
        "f_min": 10.0,
        "f_max": 2048.0,
        "delta_f": 0.125,
    }
    domain = build_domain(domain_settings)
    intrinsic_dict = {
        "mass_1": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
        "mass_2": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
        "mass_ratio": "bilby.gw.prior.UniformInComponentsMassRatio(minimum=0.125, maximum=1.0)",
        "chirp_mass": "bilby.gw.prior.UniformInComponentsChirpMass(minimum=25.0, maximum=100.0)",
        "luminosity_distance": 1000.0,
        "theta_jn": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
        "phase": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
        "a_1": "bilby.core.prior.Uniform(minimum=0.0, maximum=0.99)",
        "a_2": "bilby.core.prior.Uniform(minimum=0.0, maximum=0.99)",
        "tilt_1": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
        "tilt_2": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
        "phi_12": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
        "phi_jl": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
        "geocent_time": 0.0,
    }
    prior = build_prior_with_defaults(intrinsic_dict)
    p = prior.sample()
    p = {
        "mass_ratio": 0.3501852584069329,
        "chirp_mass": 31.709276525188667,
        "luminosity_distance": 1000.0,
        "theta_jn": 1.3663250108421872,
        "phase": 2.3133395191342094,
        "a_1": 0.9082488389607664,
        "a_2": 0.23195443013657285,
        "tilt_1": 2.2991912365076708,
        "tilt_2": 2.2878677821511086,
        "phi_12": 2.3726027637572384,
        "phi_jl": 1.5356479043406908,
        "geocent_time": 0.0,
    }

    wfg = WaveformGenerator(
        # "SEOBNRv4PHM",
        "IMRPhenomXPHM",
        domain,
        20.0,
        f_start=10.0,
        spin_conversion_phase=0.0,
    )

    pol_m = wfg.generate_hplus_hcross_m(p)

    phase_shift = np.random.uniform(high=2 * np.pi)
    print(f"{phase_shift:.2f}")
    pol = sum_contributions_m(pol_m, phase_shift=phase_shift)

    pol_ref = wfg.generate_hplus_hcross({**p, "phase": p["phase"] + phase_shift})
    # m = mismatch(
    #     apply_frequency_mask(pol, wfg.domain), apply_frequency_mask(pol_ref, wfg.domain)
    # )
    # print(f"mismatch {m:.1e}")

    import matplotlib.pyplot as plt

    x = wfg.domain()
    plt.xlim((10, 512))
    plt.xscale("log")
    plt.plot(x, pol_ref["h_plus"].real)
    plt.plot(x, pol["h_plus"].real)
    plt.plot(x, (pol_ref["h_plus"] - pol["h_plus"]).real)
    plt.show()



import numpy as np


def get_version():
    try:
        from dingo import __version__

        return __version__
    except ImportError:
        return None


def recursive_check_dicts_are_equal(dict_a, dict_b):
    if dict_a.keys() != dict_b.keys():
        return False
    else:
        for k, v_a in dict_a.items():
            v_b = dict_b[k]
            if type(v_a) != type(v_b):
                return False
            if type(v_a) == dict:
                if not recursive_check_dicts_are_equal(v_a, v_b):
                    return False
            elif not np.all(v_a == v_b):
                return False
    return True



import argparse
from pathlib import Path

import h5py
import torch
import yaml
import json
from pprint import pprint

from dingo.core.dataset import DingoDataset
from dingo.gw.SVD import SVDBasis
from dingo.core.result import Result
from dingo.gw.dataset import WaveformDataset
from dingo.gw.noise.asd_dataset import ASDDataset


def ls():
    parser = argparse.ArgumentParser()
    parser.add_argument("file_name", type=str)
    args = parser.parse_args()

    path = Path(args.file_name)
    if path.suffix == ".pt":
        print("Extracting information about torch model.\n")
        d = torch.load(path, map_location=torch.device("cpu"))
        print(f"Version: {d.get('version')}\n")
        print(f"Model epoch: {d['epoch']}\n")
        print("Model metadata:")
        print(
            yaml.dump(
                d["metadata"],
                default_flow_style=False,
                sort_keys=False,
            )
        )

    elif path.suffix == ".hdf5":

        dataset_type = determine_dataset_type(args.file_name)

        if dataset_type == "gw_result" or dataset_type == "core_result":
            result = Result(file_name=args.file_name)
            print(f"Version: {result.version}")
            print("\nDingo Result\n" + "============\n")

            print(
                "Metadata\n"
                + "--------\n"
                + yaml.dump(
                    result.settings,
                    default_flow_style=False,
                    sort_keys=False,
                )
            )
            if result.event_metadata:
                print(
                    "Event information:\n"
                    + "------------------\n"
                    + yaml.dump(
                        result.event_metadata,
                        default_flow_style=False,
                        sort_keys=False,
                    ),
                )
            if result.importance_sampling_metadata is not None:
                print(
                    "Importance sampling:\n"
                    + "--------------------\n"
                    + yaml.dump(
                        result.importance_sampling_metadata,
                        default_flow_style=False,
                        sort_keys=False,
                    ),
                )
            if result.log_evidence:
                print("Summary:\n" + "--------")
                result.print_summary()

        elif dataset_type == "svd_basis":
            svd = SVDBasis(file_name=args.file_name)
            print(f"Dingo version: {svd.version}")
            print("\nSVD Basis\n" + "=========\n")

            print(f"Basis size: {svd.n}.")
            print("\nValidation summary:\n" + "-------------------")
            svd.print_validation_summary()

        elif dataset_type == "waveform_dataset":
            waveform_dataset = WaveformDataset(file_name=args.file_name)
            print(f"Dingo version: {waveform_dataset.version}")
            print("\nWaveform dataset\n" + "================\n")

            print(f"Dataset size: {len(waveform_dataset)}")

            print(
                "\nSettings\n"
                + "--------\n"
                + yaml.dump(
                    waveform_dataset.settings,
                    default_flow_style=False,
                    sort_keys=False,
                )
            )

            if waveform_dataset.svd:
                svd = SVDBasis(dictionary=waveform_dataset.svd)
                print("\nSVD validation summary:\n" + "---------------------------")
                svd.print_validation_summary()

        elif dataset_type == "asd_dataset":
            asd_dataset = ASDDataset(file_name=args.file_name)
            print(f"Dingo version: {asd_dataset.version}")
            print("\nASD dataset\n" + "================\n")

            print(f"Dataset size: {asd_dataset.length_info}\n")
            print(f"GPS times (min/max): {asd_dataset.gps_info}")

            print(
                "\nSettings\n"
                + "--------\n"
                + yaml.dump(
                    asd_dataset.settings,
                    default_flow_style=False,
                    sort_keys=False,
                )
            )

        elif dataset_type == "trained_model":
            with h5py.File(args.file_name, "r") as f:
                print("Extracting information about torch model.\n")
                print(f"Version: {f.attrs['version']}")
                print(f"Model epoch: {f.attrs['epoch']}")
                print("Model metadata:")

                for d in ['model_kwargs', 'metadata']:
                    json_data = json.loads(f['serialized_dicts'][d][()])
                    print(f"\n{d}:\n" + "-"*(len(d)+1))
                    pprint(json_data)

        else:
            # Legacy (before dataset_type identifier).
            try:
                svd = SVDBasis(file_name=args.file_name)
                print(f"SVD dataset of size n={svd.n}.")
                print("Validation summary:")
                svd.print_validation_summary()

            except KeyError:

                dataset = DingoDataset(
                    file_name=args.file_name,
                    data_keys=[
                        "svd",
                    ],
                )
                if dataset.settings is not None:
                    print(
                        yaml.dump(
                            dataset.settings,
                            default_flow_style=False,
                            sort_keys=False,
                        )
                    )
                if dataset.svd is not None:
                    svd = SVDBasis(dictionary=dataset.svd)
                    print("SVD validation summary:")
                    svd.print_validation_summary()

    elif path.suffix == ".yaml":
        with open(path, "r") as f:
            settings = yaml.safe_load(f)
        print(
            yaml.dump(
                settings,
                default_flow_style=False,
                sort_keys=False,
            )
        )

    else:
        print("File type unrecognized.")


def determine_dataset_type(file_name):
    with h5py.File(file_name, "r") as f:
        if "dataset_type" in f.attrs:
            return f.attrs["dataset_type"]
        else:
            return None



from typing import Optional

import torch
import numpy as np
import argparse
import os
from os.path import dirname, join
from pathlib import Path
import yaml

from dingo.core.models import PosteriorModel
from dingo.core.utils.plotting import plot_corner_multi
from dingo.gw.data.event_dataset import EventDataset
from dingo.gw.inference.gw_samplers import GWSampler, GWSamplerGNPE
from dingo.gw.data.data_preparation import get_event_data_and_domain, \
    parse_settings_for_raw_data
from dingo.gw.inference.visualization import load_ref_samples


def parse_args():
    parser = argparse.ArgumentParser(
        description="Infer the posterior of a GW event using a trained dingo model.",
    )
    parser.add_argument(
        "--out_directory",
        type=str,
        default=None,
        help="Directory for output of analysis (samples, plots).",
    )
    parser.add_argument(
        "--model",
        type=str,
        help="Path to trained dingo model.",
    )
    parser.add_argument(
        "--model_init",
        type=str,
        default=None,
        help="Path to trained dingo model for initialization of gnpe. Only required if "
        "gnpe is used.",
    )
    parser.add_argument(
        # TODO: this should be renamed --events or similar
        "--gps_time_event",
        type=str,
        required=True,
        nargs="+",
        help="List of GPS times of the events to be analyzed. Used to download the GW "
        "event data, or search for it in the dataset file. Can also be a string with "
        "the path to a file containing the event data.",
    )
    parser.add_argument(
        "--event_dataset",
        type=str,
        default=None,
        help="Path to dataset for GW event data. If set, GW event data is cached in "
        "this dataset, such that it only needs to be downloaded once.",
    )
    parser.add_argument(
        "--event_dataset_init",
        type=str,
        default=None,
        help="Path to dataset for GW event data, for gnpe initialization model "
        "model_init. If set, GW event data is cached in this dataset, such that it "
        "only needs to be downloaded once.",
    )
    parser.add_argument(
        "--num_samples",
        type=int,
        default=50_000,
        help="Number of posterior samples per event.",
    )
    parser.add_argument(
        "--num_gnpe_iterations",
        type=int,
        default=30,
        help="Number of gnpe iterations.",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=None,
        help="Batch size for sampling.",
    )
    parser.add_argument(
        "--reference_file",
        type=str,
        default=None,
        help="File with path to reference samples (e.g., bilby, LALInference). If set, "
        "the corresponding samples are loaded and compared in a cornerplot.",
    )
    # settings for raw GW data
    parser.add_argument(
        "--time_psd",
        type=float,
        default=1024,
        help="Time in seconds used for PSD estimation.",
    )
    parser.add_argument(
        "--time_buffer",
        type=float,
        default=2.0,
        help="Buffer time in seconds. The analyzed strain segment extends up to "
        "gps_time_event + time_buffer.",
    )
    parser.add_argument(
        "--suffix",
        type=str,
        default="",
        help="Optional suffix for sample name.",
    )
    parser.add_argument(
        "--get_log_prob",
        action="store_true",
        help="If set, log_probs of samples are saved. For GNPE models, this invokes "
        "training of an unconditional density estimator to recover the density.",
    )
    parser.add_argument(
        "--save_low_latency",
        action="store_true",
        help="Only relevant for GNPE models and if args.get_log_prob is set. In that "
        "case, and if save_low_latency is set, intermediate results without log_prob "
        "are saved.",
    )
    parser.add_argument(
        "--density_settings",
        type=str,
        default=None,
        help="Path to yaml file with settings for density estimation. Only used if "
        "log_prob is requested for a gnpe model.",
    )
    parser.add_argument(
        "--exit_command",
        type=str,
        default=None,
        nargs="+",
        help="If set, run os.system(args.exit_command) before exiting.",
    )

    args = parser.parse_args()

    return args


def get_event_data(event, args, model, ref=None):
    # if event is a float: interpret it as gps time, download the data
    # else: interpret it as path to file with event data
    try:
        time_event = float(event)
    except ValueError:
        time_event = None

    if time_event is not None:
        # get raw event data, and prepare it for the network domain
        event_data, _ = get_event_data_and_domain(
            model.metadata,
            time_event,
            args.time_psd,
            args.time_buffer,
            args.event_dataset,
        )

        event_metadata = parse_settings_for_raw_data(model.metadata, args.time_psd,
                                                     args.time_buffer)

        # Put the metadata in the same format as provided by dingo_pipe data_generation.
        # (This is a bit ad hoc, should be improved.)
        event_metadata["time_event"] = time_event
        event_metadata["psd_duration"] = event_metadata.pop("time_psd")
        window = event_metadata.pop("window")
        event_metadata["T"] = window["T"]
        del event_metadata["time_segment"]
        event_metadata["window_type"] = window["type"]
        event_metadata["roll_off"] = window["roll_off"]
        event_metadata["f_min"] = model.metadata["dataset_settings"]["domain"]["f_min"]
        event_metadata["f_max"] = model.metadata["dataset_settings"]["domain"]["f_max"]

        if ref is None or time_event not in ref:
            label = f"gps-{time_event}{args.suffix}"
        else:
            name_event = ref[time_event]["event_name"]
            label = name_event + args.suffix

    else:
        # load file with event data
        event_dataset = EventDataset(file_name=event)
        event_data = event_dataset.data
        event_metadata = event_dataset.settings
        label = Path(event).stem + args.suffix

    return event_data, event_metadata, label


def prepare_log_prob(
    sampler,
    num_samples: int,
    nde_settings: dict,
    batch_size: Optional[int] = None,
    threshold_std: Optional[float] = np.inf,
    remove_init_outliers: Optional[float] = 0.0,
    low_latency_label: str = None,
    outdir: str = None,
):
    """
    Prepare gnpe sampling with log_prob. This is required, since in its vanilla
    form gnpe does not provide the density for its samples.

    Specifically, we train an unconditional neural density estimator (nde) for the
    gnpe proxies. This requires running the gnpe sampler till convergence, and
    extracting the gnpe proxies after the final gnpe iteration. The nde is trained
    to match the distribution over gnpe proxies, which provides a way of rapidly
    sampling (converged!) gnpe proxies *and* evaluating the log_prob.

    After this preparation step, self.run_sampler can leverage
    self.gnpe_proxy_sampler (which is based on the aforementioned trained nde) to
    sample gnpe proxies, such that one gnpe iteration is sufficient. The
    log_prob of
    the samples in the *joint* space (inference parameters + gnpe proxies) is then
    simply given by the sum of the corresponding log_probs (from self.model and
    self.gnpe_proxy_sampler.model).

    Parameters
    ----------
    num_samples: int
        number of samples for training of nde
    batch_size: int = None
        batch size for sampler
    threshold_std: float = np.inf
        gnpe proxies deviating by more then threshold_std standard deviations from
        the proxy mean (along any axis) are discarded.
    low_latency_label: str = None
        File label for low latency samples (= samples used for training nde).
        If None, these samples are not saved.
    outdir: str = None
        Directory in which low latency samples are saved. Needs to be set if
        low_latency_label is not None.
    """
    sampler.remove_init_outliers = remove_init_outliers
    sampler.run_sampler(num_samples, batch_size)
    if low_latency_label is not None:
        sampler.to_hdf5(label=low_latency_label, outdir=outdir)
    result = sampler.to_result()
    nde_settings["training"]["device"] = str(sampler.model.device)
    unconditional_model = result.train_unconditional_flow(
        sampler.gnpe_proxy_parameters,
        nde_settings,
        threshold_std=threshold_std,
    )

    # Prepare sampler with unconditional model as initialization. This should only use
    # one iteration and also not remove any outliers.
    sampler.init_sampler = GWSampler(model=unconditional_model)
    sampler.num_iterations = 1
    sampler.remove_init_outliers = 0.0  # Turn off for final sampler.


def analyze_event():
    args = parse_args()

    if torch.cuda.is_available():
        device = "cuda"
    else:
        device = "cpu"

    model = PosteriorModel(args.model, device=device, load_training_info=False)
    epoch = model.epoch
    wf_model = model.metadata["dataset_settings"]["waveform_generator"]["approximant"]

    # create analysis directory
    if args.out_directory is None:
        args.out_directory = join(dirname(args.model), f"analysis_epoch-{epoch}")
    os.makedirs(args.out_directory, exist_ok=True)

    #
    if args.reference_file is not None:
        with open(args.reference_file, "r") as f:
            ref = yaml.safe_load(f)[wf_model]
    else:
        ref = None

    if args.model_init is not None:
        gnpe = True
        init_model = PosteriorModel(
            args.model_init, device=device, load_training_info=False
        )
        init_sampler = GWSampler(model=init_model)
        sampler = GWSamplerGNPE(
            model=model,
            init_sampler=init_sampler,
            num_iterations=args.num_gnpe_iterations,
        )
    else:
        gnpe = False
        sampler = GWSampler(model=model)

    # sample posterior for events
    for time_event in args.gps_time_event:
        print(f"Analyzing event at {time_event}.")

        event_data, event_metadata, label = get_event_data(time_event, args, model, ref)

        sampler.context = event_data
        sampler.event_metadata = event_metadata

        if gnpe and args.get_log_prob:
            # GNPE generally does not provide straightforward access to the log_prob.
            # If requested, need to train an initialization model for the GNPE proxies.
            with open(args.density_settings) as f:
                density_settings = yaml.safe_load(f)
            if args.save_low_latency:
                low_latency_label = label + "_low-latency"
            else:
                low_latency_label = None
            prepare_log_prob(
                sampler,
                batch_size=args.batch_size,
                low_latency_label=low_latency_label,
                outdir=args.out_directory,
                **density_settings,
            )
            if low_latency_label is not None and sampler.iteration_tracker.store_data:
                np.save(
                    join(
                        args.out_directory,
                        f"gnpe_trajectories_{low_latency_label}.npy",
                    ),
                    sampler.iteration_tracker.data,
                )

        sampler.run_sampler(
            args.num_samples,
            batch_size=args.batch_size,
        )

        # if no reference samples are available, simply save the dingo samples
        if ref is None or time_event not in ref:
            sampler.to_hdf5(label=label, outdir=args.out_directory)

        # if reference samples are available, save dingo samples and additionally
        # compare to the reference method in a corner plot
        else:
            ref_samples_file = ref[time_event]["reference_samples"]["file"]
            ref_method = ref[time_event]["reference_samples"]["method"]
            sampler.to_hdf5(label=label, outdir=args.out_directory)
            ref_samples = load_ref_samples(ref_samples_file)
            plot_corner_multi(
                [ref_samples, sampler.samples],
                labels=[ref_method, "Dingo"],
                filename=join(args.out_directory, f"cornerplot_{label}.pdf")
            )
    if args.exit_command:
        os.system(" ".join(args.exit_command))


if __name__ == "__main__":
    analyze_event()



from .detector_transforms import *
from .noise_transforms import *
from .parameter_transforms import *
from .general_transforms import *
from .gnpe_transforms import *
from .inference_transforms import *


import numpy as np
import pytest
from dingo.gw.prior import BBHExtrinsicPriorDict, BBHPriorDict, default_extrinsic_dict, default_intrinsic_dict


def test_prior_constraint():
    dict = {
        'mass_1': 'bilby.core.prior.Uniform(minimum=10.0, maximum=80.0)',
        'mass_2': 'bilby.core.prior.Uniform(minimum=10.0, maximum=80.0)',
        'mass_ratio': 'bilby.core.prior.Constraint(minimum=0.125, maximum=1.0)',
    }
    prior = BBHPriorDict(dict)
    samples = prior.sample(1000)
    assert np.all(samples['mass_1'] > samples['mass_2'])


def test_mean_std():
    num_samples = 100000
    eps = 0.01
    keys = ['ra', 'dec', 'luminosity_distance']
    prior = BBHExtrinsicPriorDict(default_extrinsic_dict)
    mean_exact, std_exact = prior.mean_std(keys)
    mean_approx, std_approx = prior.mean_std(keys,
                                             sample_size=num_samples,
                                             force_numerical=True)
    ratios_exact = (np.array(list(mean_exact.values())) /
                    np.array(list(std_exact.values())))
    ratios_approx = (np.array(list(mean_approx.values())) /
                     np.array(list(std_approx.values())))
    assert list(mean_exact.keys()) == keys
    assert np.allclose(ratios_exact, ratios_approx,
                       atol=eps, rtol=eps)



# Transformations between PE spins and Cartesian spins
import pandas as pd
import lal
import lalsimulation as LS

DINGO_PE_SPIN_PARAMETERS = (
    "theta_jn",
    "phi_jl",
    "tilt_1",
    "tilt_2",
    "phi_12",
    "a_1",
    "a_2",
)
DINGO_CARTESIAN_SPIN_PARAMETERS = (
    "iota",
    "spin_1x",
    "spin_1y",
    "spin_1z",
    "spin_2x",
    "spin_2y",
    "spin_2z",
)


def component_masses(p):
    if "m1" in p and "m2" in p:
        return p["m1"], p["m2"]

    Mc = p["chirp_mass"]
    q = p["mass_ratio"]
    eta = q / (1 + q) ** 2
    M = Mc * eta ** (-3 / 5)

    m1 = M / (1 + q)
    m2 = q * m1
    return m1, m2


def cartesian_spins(p, f_ref):
    """
    Transform PE spins to cartesian spins.

    Parameters
    ----------
    p: dict
        contains parameters, including PE spins
    f_ref: float
        reference frequency for definition of spins

    Returns
    -------
    result: dict
        parameters, including cartesian spins
    """
    m1, m2 = component_masses(p)
    m1 *= lal.MSUN_SI
    m2 *= lal.MSUN_SI

    # Get pe spin params. Output is order identically to DINGO_CARTESIAN_SPIN_PARAMETERS,
    # iota, spin_1x, spin_1y, spin_1z, spin_2x, spin_2y, spin_2z.
    cart_spin_params = LS.SimInspiralTransformPrecessingNewInitialConditions(
        p["theta_jn"],
        p["phi_jl"],
        p["tilt_1"],
        p["tilt_2"],
        p["phi_12"],
        p["a_1"],
        p["a_2"],
        m1,
        m2,
        f_ref,
        p["phase"],
    )

    # remove PE spin parameters
    result = {k: v for k, v in p.items() if k not in DINGO_PE_SPIN_PARAMETERS}
    # add cartesian spin parameters
    for param_name, param_value in zip(
        DINGO_CARTESIAN_SPIN_PARAMETERS, cart_spin_params
    ):
        result[param_name] = param_value

    return result


def pe_spins(p, f_ref):
    """
    Transform cartesian spins to PE spins.

    Parameters
    ----------
    p: dict
        contains parameters, including cartesian spins
    f_ref: float
        reference frequency for definition of spins

    Returns
    -------
    result: dict
        parameters, including PE spins
    """
    m1, m2 = component_masses(p)

    # Get pe spin params. Output is order identically to DINGO_PE_SPIN_PARAMETERS,
    # theta_jn, phi_jl, tilt_1, tilt_2, phi_12, a_1, a_2.
    pe_spin_params = LS.SimInspiralTransformPrecessingWvf2PE(
        p["iota"],
        p["spin_1x"],
        p["spin_1y"],
        p["spin_1z"],
        p["spin_2x"],
        p["spin_2y"],
        p["spin_2z"],
        m1,
        m2,
        f_ref,
        p["phase"],
    )

    # remove cartesian spin parameters
    result = {k: v for k, v in p.items() if k not in DINGO_CARTESIAN_SPIN_PARAMETERS}
    # add PE spin parameters
    for param_name, param_value in zip(DINGO_PE_SPIN_PARAMETERS, pe_spin_params):
        result[param_name] = param_value

    return result


def change_spin_conversion_phase(samples, f_ref, sc_phase_old, sc_phase_new):
    """
    Change the phase used to convert cartesian spins to PE spins. The cartesian spins
    are independent of the spin conversion phase. When converting from cartesian spins
    to PE spins, the phase value has an impact on theta_jn and phi_jl.

    The usual convention for the PE spins is to use the phase parameter for the conversion
    (cart. spins <--> PE spins), but for dingo-IS with the synthetic phase extension we
    need to use another convention, where the PE spins are defined with spin conversion
    phase 0. This function transforms between the different conventions.

    Parameters
    ----------
    samples: pd.Dataframe
        Parameters.
    f_ref: float
        Reference frequency for definition of spins.
    sc_phase_old: float or None
        Spin conversion phase used for input parameters. If None, use the phase parameter.
    sc_phase_new: float or None
        Spin conversion phase used for output parameters. If None, use the phase
        parameter.

    Returns
    -------
    p_new:
        parameters with changed spin conversion phase
    """
    samples = samples.astype(float)
    if sc_phase_old == sc_phase_new:
        return samples

    # check validity of phases
    for sc_phase in [sc_phase_old, sc_phase_new]:
        if sc_phase is None:
            if "phase" not in samples:
                raise ValueError(
                    "Using phase parameter for spin conversion, but phase not in "
                    "parameters."
                )
        elif type(sc_phase) not in (int, float):
            raise ValueError(
                f"Phase for spin conversion needs to be either a number or None; got"
                f"{sc_phase}."
            )

    phase_old = sc_phase_old
    phase_new = sc_phase_new
    samples_new = {}
    for idx, sample in samples.to_dict(orient="index").items():
        try:
            if sc_phase_old is None:
                phase_old = sample["phase"]
            if sc_phase_new is None:
                phase_new = sample["phase"]

            # transform to cartesian spins (which is the fundamental basis, independent of
            # the phase) with the old spin conversion phase, and back to PE spins with the
            # new spin conversion phase
            sample_cartesian = cartesian_spins({**sample, "phase": phase_old}, f_ref)
            sample_pe_new = pe_spins({**sample_cartesian, "phase": phase_new}, f_ref)

            # The conversions above will set the phase parameter to sc_phase_new, however,
            # this is only the phase used for spin conversion, *not* the actual phase for
            # the parameters. Below, we set the phase to its correct (i.e., input) value.
            if "phase" in sample:
                sample_pe_new["phase"] = sample["phase"]
            else:
                sample_pe_new.pop("phase")

            samples_new[idx] = sample_pe_new
        except RuntimeError:
            print("Failed to convert spins. Saving sample unchanged.")
            samples_new[idx] = sample

    return pd.DataFrame.from_dict(samples_new, orient="index")



"""
Implementation of the neural spline flow (NSF). Most of this code is adapted
from the uci.py example from https://github.com/bayesiains/nsf.
"""
import copy

import torch
import torch.nn as nn
import glasflow.nflows as nflows  # nflows not maintained, so use this maintained fork
from glasflow.nflows import distributions, flows, transforms
import glasflow.nflows.nn.nets as nflows_nets
from dingo.core.utils import torchutils
from dingo.core.nn.enets import create_enet_with_projection_layer_and_dense_resnet
from typing import Union, Callable, Tuple


def create_linear_transform(param_dim: int):
    """
    Create the composite linear transform PLU.

    :param param_dim: int
        dimension of the parameter space
    :return: nde.Transform
        the linear transform PLU
    """

    return transforms.CompositeTransform(
        [
            transforms.RandomPermutation(features=param_dim),
            transforms.LULinear(param_dim, identity_init=True),
        ]
    )


def create_base_transform(
    i: int,
    param_dim: int,
    context_dim: int = None,
    hidden_dim: int = 512,
    num_transform_blocks: int = 2,
    activation: str = "relu",
    dropout_probability: float = 0.0,
    batch_norm: bool = False,
    num_bins: int = 8,
    tail_bound: float = 1.0,
    apply_unconditional_transform: bool = False,
    base_transform_type: str = "rq-coupling",
):
    """
    Build a base NSF transform of y, conditioned on x.

    This uses the PiecewiseRationalQuadraticCoupling transform or
    the MaskedPiecewiseRationalQuadraticAutoregressiveTransform, as described
    in the Neural Spline Flow paper (https://arxiv.org/abs/1906.04032).

    Code is adapted from the uci.py example from
    https://github.com/bayesiains/nsf.

    A coupling flow fixes half the components of y, and applies a transform
    to the remaining components, conditioned on the fixed components. This is
    a restricted form of an autoregressive transform, with a single split into
    fixed/transformed components.

    The transform here is a neural spline flow, where the flow is parametrized
    by a residual neural network that depends on y_fixed and x. The residual
    network consists of a sequence of two-layer fully-connected blocks.

    :param i: int
        index of transform in sequence
    :param param_dim: int
        dimensionality of y
    :param context_dim: int = None
        dimensionality of x
    :param hidden_dim: int = 512
        number of hidden units per layer
    :param num_transform_blocks: int = 2
        number of transform blocks comprising the transform
    :param activation: str = 'relu'
        activation function
    :param dropout_probability: float = 0.0
        dropout probability for regularization
    :param batch_norm: bool = False
        whether to use batch normalization
    :param num_bins: int = 8
        number of bins for the spline
    :param tail_bound: float = 1.
    :param apply_unconditional_transform: bool = False
        whether to apply an unconditional transform to fixed components
    :param base_transform_type: str = 'rq-coupling'
        type of base transform, one of {rq-coupling, rq-autoregressive}

    :return: Transform
        the NSF transform
    """

    activation_fn = torchutils.get_activation_function_from_string(activation)

    if base_transform_type == "rq-coupling":
        if param_dim == 1:
            mask = torch.tensor([1], dtype=torch.uint8)
        else:
            mask = nflows.utils.create_alternating_binary_mask(
                param_dim, even=(i % 2 == 0)
            )
        return transforms.PiecewiseRationalQuadraticCouplingTransform(
            mask=mask,
            transform_net_create_fn=(
                lambda in_features, out_features: nflows_nets.ResidualNet(
                    in_features=in_features,
                    out_features=out_features,
                    hidden_features=hidden_dim,
                    context_features=context_dim,
                    num_blocks=num_transform_blocks,
                    activation=activation_fn,
                    dropout_probability=dropout_probability,
                    use_batch_norm=batch_norm,
                )
            ),
            num_bins=num_bins,
            tails="linear",
            tail_bound=tail_bound,
            apply_unconditional_transform=apply_unconditional_transform,
        )

    elif base_transform_type == "rq-autoregressive":
        return transforms.MaskedPiecewiseRationalQuadraticAutoregressiveTransform(
            features=param_dim,
            hidden_features=hidden_dim,
            context_features=context_dim,
            num_bins=num_bins,
            tails="linear",
            tail_bound=tail_bound,
            num_blocks=num_transform_blocks,
            use_residual_blocks=True,
            random_mask=False,
            activation=activation_fn,
            dropout_probability=dropout_probability,
            use_batch_norm=batch_norm,
        )

    else:
        raise ValueError


def create_transform(
    num_flow_steps: int, param_dim: int, context_dim: int, base_transform_kwargs: dict
):
    """
    Build a sequence of NSF transforms, which maps parameters y into the
    base distribution u (noise). Transforms are conditioned on context data x.

    Note that the forward map is f^{-1}(y, x).

    Each step in the sequence consists of
        * A linear transform of y, which in particular permutes components
        * A NSF transform of y, conditioned on x.
    There is one final linear transform at the end.

    :param num_flow_steps: int,
        number of transforms in sequence
    :param param_dim: int,
        dimensionality of parameter space (y)
    :param context_dim: int,
        dimensionality of context (x)
    :param base_transform_kwargs: int
        hyperparameters for NSF step
    :return: Transform
        the NSF transform sequence
    """

    transform = transforms.CompositeTransform(
        [
            transforms.CompositeTransform(
                [
                    create_linear_transform(param_dim),
                    create_base_transform(
                        i, param_dim, context_dim=context_dim, **base_transform_kwargs
                    ),
                ]
            )
            for i in range(num_flow_steps)
        ]
        + [create_linear_transform(param_dim)]
    )

    return transform


class FlowWrapper(nn.Module):
    """
    This class wraps the neural spline flow. It is required for multiple
    reasons. (i) some embedding networks take tuples as input, which is not
    supported by the nflows package. (ii) paralellization across multiple
    GPUs requires a forward method, but the relevant flow method for training
    is log_prob.
    """

    def __init__(self, flow: flows.base.Flow, embedding_net: nn.Module = None):
        """

        :param flow: flows.base.Flow
        :param embedding_net: nn.Module
        """
        super(FlowWrapper, self).__init__()
        self.embedding_net = embedding_net
        self.flow = flow

    def log_prob(self, y, *x):
        if self.embedding_net is not None:
            x = torchutils.forward_pass_with_unpacked_tuple(self.embedding_net, x)
        if len(x) > 0:
            return self.flow.log_prob(y, x)
        else:
            # if there is no context
            return self.flow.log_prob(y)

    def sample(self, *x, num_samples=1):
        if self.embedding_net is not None:
            x = torchutils.forward_pass_with_unpacked_tuple(self.embedding_net, x)
        if len(x) > 0:
            return torch.squeeze(self.flow.sample(num_samples, x))
        else:
            # if there is no context, omit the context argument
            return torch.squeeze(self.flow.sample(num_samples))

    def sample_and_log_prob(self, *x, num_samples=1):
        if self.embedding_net is not None:
            x = torchutils.forward_pass_with_unpacked_tuple(self.embedding_net, x)
        if len(x) > 0:
            sample, log_prob = self.flow.sample_and_log_prob(num_samples, x)
            return torch.squeeze(sample), torch.squeeze(log_prob)
        else:
            # if there is no context, omit the context argument
            sample, log_prob = self.flow.sample_and_log_prob(num_samples)
            return torch.squeeze(sample), torch.squeeze(log_prob)

    def forward(self, y, *x):
        if len(x) > 0:
            return self.log_prob(y, *x)
        else:
            # if there is no context, omit the context argument
            return self.log_prob(y)


def create_nsf_model(
    input_dim: int,
    context_dim: int,
    num_flow_steps: int,
    base_transform_kwargs: dict,
    embedding_net_builder: Union[Callable, str] = None,
    embedding_net_kwargs: dict = None,
):
    """
    Build NSF model. This models the posterior distribution p(y|x).

    The model consists of
        * a base distribution (StandardNormal, dim(y))
        * a sequence of transforms, each conditioned on x

    :param input_dim: int,
        dimensionality of y
    :param context_dim: int,
        dimensionality of the (embedded) context
    :param num_flow_steps: int,
        number of sequential transforms
    :param base_transform_kwargs: dict,
        hyperparameters for transform steps
    :param embedding_net_builder: Callable=None,
        build function for embedding network TODO
    :param embedding_net_kwargs: dict=None,
        hyperparameters for embedding network
    :return: Flow
        the NSF (posterior model)
    """

    if embedding_net_builder is not None:
        embedding_net = embedding_net_builder(**embedding_net_kwargs)
    else:
        embedding_net = None

    # str(embedding_net_builder).split(' ')[1]

    distribution = distributions.StandardNormal((input_dim,))
    transform = create_transform(
        num_flow_steps, input_dim, context_dim, base_transform_kwargs
    )
    flow = flows.Flow(transform, distribution, embedding_net)

    # Store hyperparameters. This is for reconstructing model when loading from
    # saved file.

    flow.model_hyperparams = {
        "input_dim": input_dim,
        "num_flow_steps": num_flow_steps,
        "context_dim": context_dim,
        "base_transform_kwargs": base_transform_kwargs,
        "embedding_net_kwargs": embedding_net_kwargs,
    }

    return flow


def create_nsf_wrapped(**kwargs):
    """
    Wraps the NSF model in a FlowWrapper. This is required for parallel
    training, and wraps the log_prob method as a forward method.
    """
    flow = create_nsf_model(**kwargs)
    return FlowWrapper(flow)


def create_nsf_with_rb_projection_embedding_net(
    nsf_kwargs: dict,
    embedding_net_kwargs: dict,
    initial_weights: dict = None,
):
    """Builds a neural spline flow with an embedding network that consists of a
    reduced basis projection followed by a residual network. Optionally initializes the
    embedding network weights.

    Parameters
    ----------
    nsf_kwargs : dict
        kwargs for neural spline flow
    embedding_net_kwargs : dict
        kwargs for emebedding network
    initial_weights : dict
        Dictionary containing the initial weights for the SVD projection. This should
        have one key 'V_rb_list', with value a list of SVD V matrices (one for each
        detector).

    Returns
    -------
    nn.Module
        Neural spline flow model
    """
    # We copy the embedding_net_kwargs to allow an insert of V_rb_list without
    # affecting the original embedding_net_kwargs. This is because we don't want to
    # save the embedding_net_kwargs with the huge V_rb_list included. This is a bit of
    # a hack; improve setting of initial weights later.

    embedding_net_kwargs = copy.deepcopy(embedding_net_kwargs)
    if initial_weights is not None:
        embedding_net_kwargs["V_rb_list"] = initial_weights["V_rb_list"]
    elif "V_rb_list" not in embedding_net_kwargs:
        embedding_net_kwargs["V_rb_list"] = None

    embedding_net = create_enet_with_projection_layer_and_dense_resnet(
        **embedding_net_kwargs
    )
    flow = create_nsf_model(**nsf_kwargs)
    model = FlowWrapper(flow, embedding_net)
    return model


def autocomplete_model_kwargs_nsf(model_kwargs, data_sample):
    """
    Autocomplete the model kwargs from train_settings and data_sample from
    the dataloader:
    (*) set input dimension of embedding net to shape of data_sample[1]
    (*) set dimension of nsf parameter space to len(data_sample[0])
    (*) set added_context flag of embedding net if required for gnpe proxies
    (*) set context dim of nsf to output dim of embedding net + gnpe proxy dim

    :param train_settings: dict
        train settings as loaded from .yaml file
    :param data_sample: list
        Sample from dataloader (e.g., wfd[0]) used for autocomplection.
        Should be of format [parameters, GW data, gnpe_proxies], where the
        last element is only there is gnpe proxies are required.
    :return: model_kwargs: dict
        updated, autocompleted model_kwargs
    """
    # set input dims from ifo_list and domain information
    model_kwargs["embedding_net_kwargs"]["input_dims"] = list(data_sample[1].shape)
    # set dimension of parameter space of nsf
    model_kwargs["nsf_kwargs"]["input_dim"] = len(data_sample[0])
    # set added_context flag of embedding net if gnpe proxies are required
    # set context dim of nsf to output dim of embedding net + gnpe proxy dim
    try:
        gnpe_proxy_dim = len(data_sample[2])
        model_kwargs["embedding_net_kwargs"]["added_context"] = True
        model_kwargs["nsf_kwargs"]["context_dim"] = (
            model_kwargs["embedding_net_kwargs"]["output_dim"] + gnpe_proxy_dim
        )
    except IndexError:
        model_kwargs["embedding_net_kwargs"]["added_context"] = False
        model_kwargs["nsf_kwargs"]["context_dim"] = model_kwargs[
            "embedding_net_kwargs"
        ]["output_dim"]
    # return model_kwargs


if __name__ == "__main__":
    pass



import numpy as np
from bilby.gw.detector import InterferometerList
from torchvision.transforms import Compose

from dingo.gw.noise.asd_dataset import ASDDataset
from dingo.gw.domains import (
    FrequencyDomain,
    build_domain,
    build_domain_from_model_metadata,
)
from dingo.gw.gwutils import get_extrinsic_prior_dict
from dingo.gw.prior import build_prior_with_defaults, split_off_extrinsic_parameters
from dingo.gw.transforms import (
    GetDetectorTimes,
    ProjectOntoDetectors,
    WhitenAndScaleStrain,
    ApplyCalibrationUncertainty,
)
from dingo.gw.waveform_generator.waveform_generator import (
    WaveformGenerator,
    NewInterfaceWaveformGenerator,
)


class GWSignal(object):
    """
    Base class for generating gravitational wave signals in interferometers. Generates
    waveform polarizations based on provided parameters, and then projects to detectors.

    Includes option for whitening the signal based on a provided ASD.
    """

    def __init__(
        self,
        wfg_kwargs: dict,
        wfg_domain: FrequencyDomain,
        data_domain: FrequencyDomain,
        ifo_list: list,
        t_ref: float,
    ):
        """
        Parameters
        ----------
        wfg_kwargs : dict
            Waveform generator parameters [approximant, f_ref, and (optionally) f_start].
        wfg_domain : FrequencyDomain
            Domain used for waveform generation. This can potentially deviate from the
            final domain, having a wider frequency range needed for waveform generation.
        data_domain : FrequencyDomain
            Domain object for final signal.
        ifo_list : list
            Names of interferometers for projection.
        t_ref : float
            Reference time that specifies ifo locations.
        """

        self._check_domains(wfg_domain, data_domain)
        self.data_domain = data_domain

        # The waveform generator potentially has a larger frequency range than the
        # domain of the trained network / requested injection / etc. This is typically
        # the case for EOB waveforms, which require the larger range to generate
        # robustly. For this reason we have two domains.

        new_interface_flag = wfg_kwargs.get("new_interface", False)
        if new_interface_flag:
            self.waveform_generator = NewInterfaceWaveformGenerator(
                domain=wfg_domain, **wfg_kwargs
            )
        else:
            self.waveform_generator = WaveformGenerator(domain=wfg_domain, **wfg_kwargs)

        self.t_ref = t_ref
        self.ifo_list = InterferometerList(ifo_list)

        # When we set self.whiten, the projection transforms are automatically prepared.
        self._calibration_envelope = None
        self._calibration_marginalization_kwargs = None
        self.whiten = False

        self.asd = None

    @staticmethod
    def _check_domains(domain_in, domain_out):
        if domain_in.f_min > domain_out.f_min or domain_in.f_max < domain_out.f_max:
            raise ValueError(
                "Output domain is not contained within WaveformGenerator domain."
            )
        if domain_in.delta_f != domain_out.delta_f:
            raise ValueError("Domains must have same delta_f.")

    @property
    def whiten(self):
        """
        Bool specifying whether to whiten (and scale) generated signals.
        """
        return self._whiten

    @whiten.setter
    def whiten(self, value):
        self._whiten = value
        self._initialize_transform()

    @property
    def calibration_marginalization_kwargs(self):
        """
        Dictionary with the following keys:

        calibration_envelope
            Dictionary of the form {"H1": filepath, "L1": filepath, ...} with locations of
            lookup tables for the calibration uncertainty curves.

        num_calibration_nodes
            Number of nodes for the calibration model.

        num_calibration_curves
            Number of calibration curves to use in marginalization.
        """
        return self._calibration_marginalization_kwargs

    @calibration_marginalization_kwargs.setter
    def calibration_marginalization_kwargs(self, value):
        self._calibration_marginalization_kwargs = value
        self._initialize_transform()

    def _initialize_transform(self):
        transforms = [
            GetDetectorTimes(self.ifo_list, self.t_ref),
            ProjectOntoDetectors(self.ifo_list, self.data_domain, self.t_ref),
        ]
        if self.calibration_marginalization_kwargs:
            transforms.append(
                ApplyCalibrationUncertainty(
                    self.ifo_list,
                    self.data_domain,
                    **self.calibration_marginalization_kwargs,
                )
            )
        if self.whiten:
            transforms.append(WhitenAndScaleStrain(self.data_domain.noise_std))
        self.projection_transforms = Compose(transforms)

    def signal(self, theta):
        """
        Compute the GW signal for parameters theta.

        Step 1: Generate polarizations
        Step 2: Project polarizations onto detectors; optionally (depending on
        self.whiten) whiten and scale.

        Parameters
        ----------
        theta: dict
            Signal parameters. Includes intrinsic parameters to be passed to waveform
            generator, and extrinsic parameters for detector projection.

        Returns
        -------
        dict
            keys:
                waveform: GW strain signal for each detector.
                extrinsic_parameters: {}
                parameters: waveform parameters
                asd (if set): amplitude spectral density for each detector
        """
        theta_intrinsic, theta_extrinsic = split_off_extrinsic_parameters(theta)
        theta_intrinsic = {k: float(v) for k, v in theta_intrinsic.items()}

        # Step 1: generate polarizations h_plus and h_cross
        polarizations = self.waveform_generator.generate_hplus_hcross(theta_intrinsic)
        polarizations = {  # truncation, in case wfg has a larger frequency range
            k: self.data_domain.update_data(v) for k, v in polarizations.items()
        }

        # Step 2: project h_plus and h_cross onto detectors
        sample = {
            "parameters": theta_intrinsic,
            "extrinsic_parameters": theta_extrinsic,
            "waveform": polarizations,
        }

        asd = self.asd
        if asd is not None:
            sample["asds"] = asd

        return self.projection_transforms(sample)

    # It would be good to have an ASD class to handle all of this functionality,
    # namely storing ASDs from numpy arrays, from ASDDatasets, loading from files,
    # etc. For now this functionality is partially implemented here.

    def signal_m(self, theta):
        """
        Compute the GW signal for parameters theta. Same as self.signal(theta) method,
        but it does not sum the contributions of the individual modes, and instead
        returns a dict {m: pol_m for m in [-l_max,...,0,...,l_max]} where each
        contribution pol_m transforms as exp(-1j * m * phase_shift) under phase shifts.

        Step 1: Generate polarizations
        Step 2: Project polarizations onto detectors;
                optionally (depending on self.whiten) whiten and scale.

        Parameters
        ----------
        theta: dict
            Signal parameters. Includes intrinsic parameters to be passed to waveform
            generator, and extrinsic parameters for detector projection.

        Returns
        -------
        dict
            keys:
                waveform:
                    GW strain signal for each detector, with individual contributions
                    {m: pol_m for m in [-l_max,...,0,...,l_max]}
                extrinsic_parameters: {}
                parameters: waveform parameters
                asd (if set): amplitude spectral density for each detector
        """
        theta_intrinsic, theta_extrinsic = split_off_extrinsic_parameters(theta)
        theta_intrinsic = {k: float(v) for k, v in theta_intrinsic.items()}

        # Step 1: generate m-contributions to polarizations h_plus and h_cross
        pol_m = self.waveform_generator.generate_hplus_hcross_m(theta_intrinsic)
        # truncation, in case wfg has a larger frequency range
        pol_m = {
            k_m: {
                k_pol: self.data_domain.update_data(v_pol)
                for k_pol, v_pol in v_m.items()
            }
            for k_m, v_m in pol_m.items()
        }

        # Step 2: project m-contributions to h_plus and h_cross onto detectors
        sample_out = {}
        for m, pol in pol_m.items():
            sample = {
                "parameters": theta_intrinsic,
                "extrinsic_parameters": theta_extrinsic,
                "waveform": pol,
            }
            if self.asd is not None:
                sample["asds"] = self.asd
            sample_out[m] = self.projection_transforms(sample)

        return sample_out

    @property
    def asd(self):
        """
        Amplitude spectral density.

        Either a single array, a dict (for individual interferometers),
        or an ASDDataset, from which random ASDs are drawn.
        """
        if isinstance(self._asd, np.ndarray):
            asd = {ifo.name: self._asd for ifo in self.ifo_list}
        elif isinstance(self._asd, dict):
            asd = self._asd
        elif isinstance(self._asd, ASDDataset):
            asd = self._asd.sample_random_asds()
        elif self._asd is None:
            return None
        else:
            raise TypeError("Invalid ASD type.")
        asd = {
            k: self.data_domain.update_data(v, low_value=1e-20) for k, v in asd.items()
        }
        return asd

    @asd.setter
    def asd(self, asd):
        ifo_names = [ifo.name for ifo in self.ifo_list]
        if isinstance(asd, ASDDataset):
            if set(asd.asds.keys()) != set(ifo_names):
                raise KeyError("ASDDataset ifos do not match signal.")
        elif isinstance(asd, dict):
            if set(asd.keys()) != set(ifo_names):
                raise KeyError("ASD ifos do not match signal.")
        elif isinstance(asd, str):
            raise NotImplementedError(
                "Still need to implement injections with ASDs defined by file names."
            )
        self._asd = asd


class Injection(GWSignal):
    """
    Produces injections of signals (with random or specified parameters) into stationary
    Gaussian noise. Output is not whitened.
    """

    def __init__(self, prior, **gwsignal_kwargs):
        """
        Parameters
        ----------
        prior : PriorDict
            Prior used for sampling random parameters.
        gwsignal_kwargs
            Arguments to be passed to GWSignal base class.
        """
        super().__init__(**gwsignal_kwargs)
        self.prior = prior

    @classmethod
    def from_posterior_model_metadata(cls, metadata):
        """
        Instantiate an Injection based on a posterior model. The prior, waveform
        settings, etc., will all be consistent with what the model was trained with.

        Parameters
        ----------
        metadata : dict
            Dict which you can get via PosteriorModel.metadata
        """
        intrinsic_prior = metadata["dataset_settings"]["intrinsic_prior"]
        extrinsic_prior = get_extrinsic_prior_dict(
            metadata["train_settings"]["data"]["extrinsic_prior"]
        )
        prior = build_prior_with_defaults({**intrinsic_prior, **extrinsic_prior})

        return cls(
            prior=prior,
            wfg_kwargs=metadata["dataset_settings"]["waveform_generator"],
            wfg_domain=build_domain(metadata["dataset_settings"]["domain"]),
            data_domain=build_domain_from_model_metadata(metadata),
            ifo_list=metadata["train_settings"]["data"]["detectors"],
            t_ref=metadata["train_settings"]["data"]["ref_time"],
        )

    def injection(self, theta):
        """
        Generate an injection based on specified parameters.

        This is a signal + noise  consistent with the amplitude spectral density in
        self.asd. If self.asd is an ASDDataset, then it uses a random ASD from this
        dataset.

        Data are not whitened.

        Parameters
        ----------
        theta : dict
            Parameters used for injection.

        Returns
        -------
        dict
            keys:
                waveform: data (signal + noise) in each detector
                extrinsic_parameters: {}
                parameters: waveform parameters
                asd (if set): amplitude spectral density for each detector
        """
        signal = self.signal(theta)
        try:
            # Be careful to use the ASD included with the signal, since each time
            # self.asd is accessed it gives a different ASD (if using an ASD dataset).
            asd = signal["asds"]
        except KeyError:
            raise ValueError("self.asd must be set in order to produce injections.")

        if self.whiten:
            print("self.whiten was set to True. Resetting to False.")
            self.whiten = False

        data = {}
        for ifo, s in signal["waveform"].items():
            noise = (
                (np.random.randn(len(s)) + 1j * np.random.randn(len(s)))
                * asd[ifo]
                * self.data_domain.noise_std
            )
            d = s + noise
            data[ifo] = self.data_domain.update_data(d, low_value=0.0)

        signal["waveform"] = data
        return signal

    def random_injection(self):
        """
        Generate a random injection.

        This is a signal + noise  consistent with the amplitude spectral density in
        self.asd. If self.asd is an ASDDataset, then it uses a random ASD from this
        dataset.

        Data are not whitened.

        Returns
        -------
        dict
            keys:
                waveform: data (signal + noise) in each detector
                extrinsic_parameters: {}
                parameters: waveform parameters
                asd (if set): amplitude spectral density for each detector
        """
        theta = self.prior.sample()
        return self.injection(theta)



from multiprocessing import Pool
from threadpoolctl import threadpool_limits
import numpy as np
import pandas as pd


def apply_func_with_multiprocessing(
    func: callable, theta: pd.DataFrame, num_processes: int = 1
) -> np.ndarray:
    """
    Call func(theta.iloc[idx].to_dict()) with multiprocessing.

    Parameters
    ----------
    func: callable

    theta : pd.DataFrame
        Parameters with multiple rows, evaluate func for each row.
    num_processes : int
        Number of parallel processes to use.

    Returns
    -------
    result: np.ndarray
        Output array, where result[idx] = func(theta.iloc[idx].to_dict())
    """
    with threadpool_limits(limits=1, user_api="blas"):

        # Generator object for theta rows. For idx this yields row idx of
        # theta dataframe, converted to dict, ready to be passed to
        # self.log_likelihood.
        theta_generator = (d[1].to_dict() for d in theta.iterrows())

        if num_processes > 1:
            with Pool(processes=num_processes) as pool:
                result = pool.map(func, theta_generator)
        else:
            result = list(map(func, theta_generator))

    return np.array(result)



import pytest
import os.path
import numpy as np
from bilby.gw.detector import InterferometerList

from dingo.gw.transforms import (
    GetDetectorTimes,
    ProjectOntoDetectors,
    SampleExtrinsicParameters,
    time_delay_from_geocenter,
)
from dingo.gw.prior import default_extrinsic_dict
from dingo.gw.domains import build_domain


@pytest.fixture
def reference_data_research_code():
    dir = os.path.dirname(os.path.realpath(__file__))
    ref_data = np.load(os.path.join(dir, "waveform_data.npy"), allow_pickle=True).item()
    sample_in = {
        "parameters": ref_data["intrinsic_parameters"],
        "waveform": {"h_cross": ref_data["hc"], "h_plus": ref_data["hp"]},
        "extrinsic_parameters": ref_data["extrinsic_parameters"],
    }
    parameters_ref = ref_data["all_parameters"]
    h_ref = ref_data["h_d_unwhitened"]
    return sample_in, parameters_ref, h_ref


@pytest.fixture
def setup_detector_projection():
    # setup arguments
    extrinsic_prior_dict = default_extrinsic_dict
    ref_time = 1126259462.391
    domain_dict = {
        "type": "FrequencyDomain",
        "f_min": 10.0,
        "f_max": 1024.0,
        "delta_f": 0.125,
    }
    ifo_list = InterferometerList(["H1", "L1"])
    domain = build_domain(domain_dict)

    # build transformations
    sample_extrinsic_parameters = SampleExtrinsicParameters(extrinsic_prior_dict)
    get_detector_times = GetDetectorTimes(ifo_list, ref_time)
    project_onto_detectors = ProjectOntoDetectors(ifo_list, domain, ref_time)

    return sample_extrinsic_parameters, get_detector_times, project_onto_detectors


def test_detector_projection_against_research_code(
    reference_data_research_code, setup_detector_projection
):
    sample_in, parameters_ref, h_ref = reference_data_research_code
    _, get_detector_times, project_onto_detector = setup_detector_projection

    sample_out = get_detector_times(sample_in)
    sample_out["extrinsic_parameters"]["H1_time"] = parameters_ref["H1_time"]
    sample_out["extrinsic_parameters"]["L1_time"] = parameters_ref["L1_time"]
    sample_out = project_onto_detector(sample_out)

    for ifo_name in ["H1", "L1"]:
        strain = sample_out["waveform"][ifo_name]
        strain_ref = h_ref[ifo_name]
        deviation = np.abs(strain_ref - strain)
        assert np.max(deviation) / np.max(np.abs(strain)) < 5e-2


def test_time_delay_from_geocenter():
    ifo_list = InterferometerList(["H1", "L1", "V1"])
    for ifo in ifo_list:
        ra = np.random.uniform(0, 2 * np.pi, size=2)
        dec = np.random.uniform(-np.pi / 2, np.pi / 2, size=2)
        ref_time = 1126259462.4
        td = time_delay_from_geocenter(ifo, ra, dec, ref_time)
        # The vectorized version of time_delay_from_geocenter uses our own
        # implementation, whereas for floats it uses bilby. Below, we iterate through
        # the individual elements of the arrays, and compare the output of our
        # vectorized implementation to bilby outputs.
        for ra_idx, dec_idx, td_idx in zip(ra, dec, td):
            assert np.allclose(
                td_idx, time_delay_from_geocenter(ifo, ra_idx, dec_idx, ref_time)
            )



#!/usr/bin/env python
""" Script to sample from a Dingo model. Based on bilby_pipe data analysis script. """
import sys
from pathlib import Path

from bilby_pipe.input import Input
from bilby_pipe.utils import parse_args, logger, convert_string_to_dict

from dingo.core.models import PosteriorModel
from dingo.gw.data.event_dataset import EventDataset
from dingo.gw.inference.gw_samplers import GWSampler, GWSamplerGNPE
from dingo.gw.inference.inference_pipeline import prepare_log_prob
from dingo.pipe.default_settings import DENSITY_RECOVERY_SETTINGS
from dingo.pipe.parser import create_parser

logger.name = "dingo_pipe"


class SamplingInput(Input):
    def __init__(self, args, unknown_args):
        super().__init__(args, unknown_args)

        # Generic initialisation
        self.meta_data = dict()
        self.result = None

        # Admin arguments
        self.ini = args.ini
        self.scheduler = args.scheduler
        # self.periodic_restart_time = args.periodic_restart_time
        self.request_cpus = args.request_cpus
        self.n_parallel = args.n_parallel

        # Naming arguments
        self.outdir = args.outdir
        self.label = args.label
        self.result_format = args.result_format

        # Event file to run on
        self.event_data_file = args.event_data_file

        # Choices for running
        self.detectors = args.detectors
        self.model = args.model
        self.model_init = args.model_init
        self.recover_log_prob = args.recover_log_prob
        self.device = args.device
        self.num_gnpe_iterations = args.num_gnpe_iterations
        self.num_samples = args.num_samples
        self.batch_size = args.batch_size
        self.density_recovery_settings = args.density_recovery_settings

        # self.sampler = args.sampler
        # self.sampler_kwargs = args.sampler_kwargs
        # self.sampling_seed = args.sampling_seed

        # Frequencies
        # self.sampling_frequency = args.sampling_frequency
        # self.minimum_frequency = args.minimum_frequency
        # self.maximum_frequency = args.maximum_frequency
        # self.reference_frequency = args.reference_frequency

        # # Waveform, source model and likelihood
        # self.waveform_generator_class = args.waveform_generator
        # self.waveform_approximant = args.waveform_approximant
        # self.catch_waveform_errors = args.catch_waveform_errors
        # self.pn_spin_order = args.pn_spin_order
        # self.pn_tidal_order = args.pn_tidal_order
        # self.pn_phase_order = args.pn_phase_order
        # self.pn_amplitude_order = args.pn_amplitude_order
        # self.mode_array = args.mode_array
        # self.waveform_arguments_dict = args.waveform_arguments_dict
        # self.numerical_relativity_file = args.numerical_relativity_file
        # self.frequency_domain_source_model = args.frequency_domain_source_model
        # self.conversion_function = args.conversion_function
        # self.generation_function = args.generation_function
        # self.likelihood_type = args.likelihood_type
        # self.reference_frame = args.reference_frame
        # self.time_reference = args.time_reference
        # self.extra_likelihood_kwargs = args.extra_likelihood_kwargs
        # self.enforce_signal_duration = args.enforce_signal_duration
        #
        # # ROQ
        # self.roq_folder = args.roq_folder
        # self.roq_scale_factor = args.roq_scale_factor
        #
        # # Calibration
        # self.calibration_model = args.calibration_model
        # self.spline_calibration_nodes = args.spline_calibration_nodes
        # self.spline_calibration_envelope_dict = args.spline_calibration_envelope_dict

        # # Marginalization
        # self.distance_marginalization = args.distance_marginalization
        # self.distance_marginalization_lookup_table = None
        # self.phase_marginalization = args.phase_marginalization
        # self.time_marginalization = args.time_marginalization
        # self.jitter_time = args.jitter_time

        self._load_event()
        self._load_sampler()

    def _load_event(self):
        event_dataset = EventDataset(file_name=self.event_data_file)
        self.context = event_dataset.data
        self.event_metadata = event_dataset.settings

    def _load_sampler(self):
        """Load the sampler and set its context based on event data."""
        model = PosteriorModel(self.model, device=self.device, load_training_info=False)

        if self.model_init is not None:
            self.gnpe = True
            init_model = PosteriorModel(
                self.model_init, device=self.device, load_training_info=False
            )
            init_sampler = GWSampler(model=init_model)
            self.dingo_sampler = GWSamplerGNPE(
                model=model,
                init_sampler=init_sampler,
                num_iterations=self.num_gnpe_iterations,
            )

        else:
            self.gnpe = False
            self.dingo_sampler = GWSampler(model=model)

        self.dingo_sampler.context = self.context
        self.dingo_sampler.event_metadata = self.event_metadata

    @property
    def density_recovery_settings(self):
        return self._density_recovery_settings

    @density_recovery_settings.setter
    def density_recovery_settings(self, settings):
        if self.recover_log_prob:
            self._density_recovery_settings = DENSITY_RECOVERY_SETTINGS[
                "ProxyRecoveryDefault"
            ]
        else:
            self._density_recovery_settings = dict()

        if settings is not None:
            if settings.lower() in ["default", "proxyrecoverydefault"]:
                self._density_recovery_settings.update(
                    DENSITY_RECOVERY_SETTINGS["ProxyRecoveryDefault"]
                )
            else:
                self._density_recovery_settings.update(convert_string_to_dict(settings))

    # @property
    # def result_directory(self):
    #     result_dir = os.path.join(self.outdir, "result")
    #     return os.path.relpath(result_dir)

    def run_sampler(self):

        if self.gnpe and self.recover_log_prob:
            logger.info(
                "GNPE network does not provide log probability. Generating "
                "samples and training a new network to recover it."
            )

            # Note that this will not save any low latency samples at present.
            prepare_log_prob(
                self.dingo_sampler,
                batch_size=self.batch_size,
                **self.density_recovery_settings,
            )

        self.dingo_sampler.run_sampler(self.num_samples, batch_size=self.batch_size)
        self.dingo_sampler.to_hdf5(label=self.label, outdir=self.result_directory)

        if self.n_parallel > 1:
            logger.info(f"Splitting Result into {self.n_parallel} parts.")
            result = self.dingo_sampler.to_result()
            sub_results = result.split(self.n_parallel)
            outdir = Path(self.result_directory)
            outdir.mkdir(parents=True, exist_ok=True)
            for n, r in enumerate(sub_results):
                file_name = self.label + f"_part{n}.hdf5"
                r.to_file(file_name=outdir / file_name)


def create_sampling_parser():
    """Data analysis parser creation"""
    return create_parser(top_level=False)


def main():
    """Data analysis main logic"""
    args, unknown_args = parse_args(sys.argv[1:], create_sampling_parser())
    # log_version_information()
    analysis = SamplingInput(args, unknown_args)
    analysis.run_sampler()
    sys.exit(0)
